#!/usr/bin/env jmxsh-R5.jar
#ident  "@(#)jmxstat.tcl        0.1.10   2017-05-23 abs"
#******************************************************************************
# $Id: jmxstat.tcl 79 2017-05-23 20:49:24Z asmith69 $
# $Revision: 79 $
# $Author: asmith69 $
# $Date: 2017-05-23 21:49:24 +0100 (Tue, 23 May 2017) $
# $HeadURL: http://localhost:18080/svn/xymon_ext/trunk/jmxstat.tcl $
#******************************************************************************
# tcl script to collect information via JMX from any J2EE server using jmxsh.jar
#
#   0.0.0 2010-11-28 abs      test version
#   0.0.1 2011-02-27 abs      functional version
#   0.0.2 2011-02-28 abs      multiple platform version
#   0.0.3 2011-03-03 abs      formatting to mimic the old jmxstat completed
#   0.0.4 2011-03-07 abs      jmxstat report started
#   0.0.5 2011-03-14 abs      jmxstat report completed
#   0.0.6 2011-03-20 abs      fix when jmx unavailable
#   0.0.7 2011-03-21 abs      fix when server down, adjustmoents for BBWin
#         2011-03-22 abs      fix for rrdtool misinterpretting date header
#         2011-03-22 abs      fix for memory pool names containing []
#         2011-03-22 abs      fix for garbage collector names containing space
#   0.0.8 2011-03-22 abs      fix for uptime processing
#   0.1.0 2011-03-22 abs      first version pushed to prod
#   0.1.1 2011-07-03 abs      dont report old gcinfo alerts
#         2011-07-08 abs      hack for gc collectors that lie about number
#         2011-07-08 abs      of collections.  Corrections for integer
#   0.1.2 2011-07-08 abs      arithmetic in Heap sizes
#         2011-11-11 abs      support for Xymon 4.3
#         2011-11-11 abs      proper flow control around down/unresponsive
#   0.1.3 2011-11-11 abs      hosts, better error checking around jmx_get
#         2012-03-07 abs      initialise JavaInstances
#         2012-07-24 pk       declare gcs variable in format_gc_body
#   0.1.4 2012-09-15 abs      better debugging
#   0.1.5 2013-03-07 pk       put \" wrapper around mbean args to jmx_get
#         2013-03-07 abs      cater for jvm not providing maxSpareThreads
#   0.1.6 2014-04-08 abs      arithmetic in GCInfo sizes
#   0.1.7 2014-08-07 abs      pre and post exec options to allow writing your
#                             own tcl - permits retrieval of java objects and
#                             provides mechanism for subsequent transformation
#                             of result.
#   0.1.8 2015-01-09 abs      re-work of 0.1.5 using MBEAN and ATTROP instead
#                             which is more reliable.
#                             new keyword "conntest" added to ini file to set
#                             test to be queried, defaults to "conn"
#         2015-02-28 abs      jmxstat runtime rrd info sent
#   0.1.9 2016-09-07 abs      XYMONNETWORK removed for xymongrep
#   0.1.10 2017-05-23 abs     portability fix for AIX
#******************************************************************************
set version "0.1.9"

###############################################################################
# IMPORTANT
# The following packages from TcLib are used:
#
# matrix.tcl Implementation of a matrix data structure for Tcl.
# ini.tcl Querying and modifying old-style windows configuration files (.ini)
# cmdline.tcl This package provides a utility for parsing command line
# textutil.tcl specifically the split.tcl code
#
# These could be provided in a generic way from TcLib, but would involve
# too many dependancies since Tcl is not necessarily installed on the platform
# let alone TcLib.
#
# Since the packages are standalone tcl code fragments they could be just as
# easily be sourced into this script, but again, providing a portable place
# for these to be located will cause unnecessary complication.
#
# Instead, the current version of these packages are included directly in
# this script, all copyrights and license information has been preserved.
#
# The code for jmxsh starts after the line located by the string JMXSTAT
#
###############################################################################
#source matrix.tcl
# matrix.tcl --
#
#	Implementation of a matrix data structure for Tcl.
#
# Copyright (c) 2001-2004 by Andreas Kupries <andreas_kupries@users.sourceforge.net>
#
# Heapsort code Copyright (c) 2003 by Edwin A. Suominen <ed@eepatents.com>,
# based on concepts in "Introduction to Algorithms" by Thomas H. Cormen et al.
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Import: matrix.tcl,v 1.23 2008/02/20 00:39:39 andreas_kupries Exp $


namespace eval ::struct {}

namespace eval ::struct::matrix {
    # Data storage in the matrix module
    # -------------------------------
    #
    # One namespace per object, containing
    #
    # - Two scalar variables containing the current number of rows and columns.
    # - Four array variables containing the array data, the caches for
    #   row heights and column widths and the information about linked arrays.
    #
    # The variables are
    # - columns #columns in data
    # - rows    #rows in data
    # - data    cell contents
    # - colw    cache of column widths
    # - rowh    cache of row heights
    # - link    information about linked arrays
    # - lock    boolean flag to disable MatTraceIn while in MatTraceOut [#532783]
    # - unset   string used to convey information about 'unset' traces from MatTraceIn to MatTraceOut.

    # counter is used to give a unique name for unnamed matrices
    variable counter 0

    # Only export one command, the one used to instantiate a new matrix
    namespace export matrix
}

# ::struct::matrix::matrix --
#
#	Create a new matrix with a given name; if no name is given, use
#	matrixX, where X is a number.
#
# Arguments:
#	name	Optional name of the matrix; if null or not given, generate one.
#
# Results:
#	name	Name of the matrix created

proc ::struct::matrix::matrix {args} {
    variable counter
    
    set src     {}
    set srctype {}

    switch -exact -- [llength [info level 0]] {
	1 {
	    # Missing name, generate one.
	    incr counter
	    set name "matrix${counter}"
	}
	2 {
	    # Standard call. New empty matrix.
	    set name [lindex $args 0]
	}
	4 {
	    # Copy construction.
	    foreach {name as src} $args break
	    switch -exact -- $as {
		= - := - as {
		    set srctype matrix
		}
		deserialize {
		    set srctype serial
		}
		default {
		    return -code error \
			    "wrong # args: should be \"matrix ?name ?=|:=|as|deserialize source??\""
		}
	    }
	}
	default {
	    # Error.
	    return -code error \
		    "wrong # args: should be \"matrix ?name ?=|:=|as|deserialize source??\""
	}
    }

    # FIRST, qualify the name.
    if {![string match "::*" $name]} {
        # Get caller's namespace; append :: if not global namespace.
        set ns [uplevel 1 [list namespace current]]
        if {"::" != $ns} {
            append ns "::"
        }
        set name "$ns$name"
    }

    if { [llength [info commands $name]] } {
	return -code error "command \"$name\" already exists, unable to create matrix"
    }

    # Set up the namespace
    namespace eval $name {
	variable columns 0
	variable rows    0

	variable data
	variable colw
	variable rowh
	variable link
	variable lock
	variable unset

	array set data  {}
	array set colw  {}
	array set rowh  {}
	array set link  {}
	set       lock  0
	set       unset {}
    }

    # Create the command to manipulate the matrix
    interp alias {} $name {} ::struct::matrix::MatrixProc $name

    # Automatic execution of assignment if a source
    # is present.
    if {$src != {}} {
	switch -exact -- $srctype {
	    matrix {_= $name $src}
	    serial {_deserialize $name $src}
	    default {
		return -code error \
			"Internal error, illegal srctype \"$srctype\""
	    }
	}
    }
    return $name
}

##########################
# Private functions follow

# ::struct::matrix::MatrixProc --
#
#	Command that processes all matrix object commands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand to invoke.
#	args	Arguments for subcommand.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::MatrixProc {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub _$cmd
    if {[llength [info commands ::struct::matrix::$sub]] == 0} {
	set optlist [lsort [info commands ::struct::matrix::_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    if {[string match __* $p]} {continue}
	    lappend xlist [string range $p 1 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_= --
#
#	Assignment operator. Copies the source matrix into the
#       destination, destroying the original information.
#
# Arguments:
#	name	Name of the matrix object we are copying into.
#	source	Name of the matrix object providing us with the
#		data to copy.
#
# Results:
#	Nothing.

proc ::struct::matrix::_= {name source} {
    _deserialize $name [$source serialize]
    return
}

# ::struct::matrix::_--> --
#
#	Reverse assignment operator. Copies this matrix into the
#       destination, destroying the original information.
#
# Arguments:
#	name	Name of the matrix object to copy
#	dest	Name of the matrix object we are copying to.
#
# Results:
#	Nothing.

proc ::struct::matrix::_--> {name dest} {
    $dest deserialize [_serialize $name]
    return
}

# ::struct::matrix::_add --
#
#	Command that processes all 'add' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'add' to invoke.
#	args	Arguments for subcommand of 'add'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_add {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name add option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __add_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__add_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 6 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_delete --
#
#	Command that processes all 'delete' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'delete' to invoke.
#	args	Arguments for subcommand of 'delete'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_delete {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name delete option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __delete_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__delete_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 9 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_format --
#
#	Command that processes all 'format' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'format' to invoke.
#	args	Arguments for subcommand of 'format'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_format {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name format option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __format_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__format_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 9 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_get --
#
#	Command that processes all 'get' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'get' to invoke.
#	args	Arguments for subcommand of 'get'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_get {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name get option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __get_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__get_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 6 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_insert --
#
#	Command that processes all 'insert' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'insert' to invoke.
#	args	Arguments for subcommand of 'insert'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_insert {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name insert option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __insert_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__insert_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 9 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_search --
#
#	Command that processes all 'search' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	args	Arguments for search.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_search {name args} {
    set mode   exact
    set nocase 0

    while {1} {
	switch -glob -- [lindex $args 0] {
	    -exact - -glob - -regexp {
		set mode [string range [lindex $args 0] 1 end]
		set args [lrange $args 1 end]
	    }
	    -nocase {
		set nocase 1
		set args [lrange $args 1 end]
	    }
	    -* {
		return -code error \
			"invalid option \"[lindex $args 0]\":\
			should be -nocase, -exact, -glob, or -regexp"
	    }
	    default {
		break
	    }
	}
    }

    # Possible argument signatures after option processing
    #
    # \ | args
    # --+--------------------------------------------------------
    # 2 | all pattern
    # 3 | row row pattern, column col pattern
    # 6 | rect ctl rtl cbr rbr pattern
    #
    # All range specifications are internally converted into a
    # rectangle.

    switch -exact -- [llength $args] {
	2 - 3 - 6 {}
	default {
	    return -code error \
		"wrong # args: should be\
		\"$name search ?option...? (all|row row|column col|rect c r c r) pattern\""
	}
    }

    set range   [lindex $args 0]
    set pattern [lindex $args end]
    set args    [lrange $args 1 end-1]

    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows

    switch -exact -- $range {
	all {
	    set ctl 0 ; set cbr $columns ; incr cbr -1
	    set rtl 0 ; set rbr $rows    ; incr rbr -1
	}
	column {
	    set ctl [ChkColumnIndex $name [lindex $args 0]]
	    set cbr $ctl
	    set rtl 0       ; set rbr $rows ; incr rbr -1
	}
	row {
	    set rtl [ChkRowIndex $name [lindex $args 0]]
	    set ctl 0    ; set cbr $columns ; incr cbr -1
	    set rbr $rtl
	}
	rect {
	    foreach {ctl rtl cbr rbr} $args break
	    set ctl [ChkColumnIndex $name $ctl]
	    set rtl [ChkRowIndex    $name $rtl]
	    set cbr [ChkColumnIndex $name $cbr]
	    set rbr [ChkRowIndex    $name $rbr]
	    if {($ctl > $cbr) || ($rtl > $rbr)} {
		return -code error "Invalid cell indices, wrong ordering"
	    }
	}
	default {
	    return -code error "invalid range spec \"$range\": should be all, column, row, or rect"
	}
    }

    if {$nocase} {
	set pattern [string tolower $pattern]
    }

    set matches [list]
    for {set r $rtl} {$r <= $rbr} {incr r} {
	for {set c $ctl} {$c <= $cbr} {incr c} {
	    set v  $data($c,$r)
	    if {$nocase} {
		set v [string tolower $v]
	    }
	    switch -exact -- $mode {
		exact  {set matched [string equal $pattern $v]}
		glob   {set matched [string match $pattern $v]}
		regexp {set matched [regexp --    $pattern $v]}
	    }
	    if {$matched} {
		lappend matches [list $c $r]
	    }
	}
    }
    return $matches
}

# ::struct::matrix::_set --
#
#	Command that processes all 'set' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'set' to invoke.
#	args	Arguments for subcommand of 'set'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_set {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name set option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __set_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__set_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 6 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::_sort --
#
#	Command that processes all 'sort' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'sort' to invoke.
#	args	Arguments for subcommand of 'sort'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_sort {name cmd args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name sort option ?arg arg ...?\""
    }
    if {[string equal $cmd "rows"]} {
	set code   r
	set byrows 1
    } elseif {[string equal $cmd "columns"]} {
	set code   c
	set byrows 0
    } else {
	return -code error \
		"bad option \"$cmd\": must be columns, or rows"
    }

    set revers 0 ;# Default: -increasing
    while {1} {
	switch -glob -- [lindex $args 0] {
	    -increasing {set revers 0}
	    -decreasing {set revers 1}
	    default {
		if {[llength $args] > 1} {
		    return -code error \
			"invalid option \"[lindex $args 0]\":\
			should be -increasing, or -decreasing"
		}
		break
	    }
	}
	set args [lrange $args 1 end]
    }
    # ASSERT: [llength $args] == 1

    if {[llength $args] != 1} {
	return -code error "wrong # args: should be \"$name sort option ?arg arg ...?\""
    }

    set key [lindex $args 0]

    if {$byrows} {
	set key [ChkColumnIndex $name $key]
	variable ${name}::rows

	# Adapted by EAS from BUILD-MAX-HEAP(A) of CRLS 6.3
	set heapSize $rows
    } else {
	set key [ChkRowIndex $name $key]
	variable ${name}::columns

	# Adapted by EAS from BUILD-MAX-HEAP(A) of CRLS 6.3
	set heapSize $columns
    }

    for {set i [expr {int($heapSize/2)-1}]} {$i>=0} {incr i -1} {
	SortMaxHeapify $name $i $key $code $heapSize $revers
    }

    # Adapted by EAS from remainder of HEAPSORT(A) of CRLS 6.4
    for {set i [expr {$heapSize-1}]} {$i>=1} {incr i -1} {
	if {$byrows} {
	    SwapRows $name 0 $i
	} else {
	    SwapColumns $name 0 $i
	}
	incr heapSize -1
	SortMaxHeapify $name 0 $key $code $heapSize $revers
    }
    return
}

# ::struct::matrix::_swap --
#
#	Command that processes all 'swap' subcommands.
#
# Arguments:
#	name	Name of the matrix object to manipulate.
#	cmd	Subcommand of 'swap' to invoke.
#	args	Arguments for subcommand of 'swap'.
#
# Results:
#	Varies based on command to perform

proc ::struct::matrix::_swap {name {cmd ""} args} {
    # Do minimal args checks here
    if { [llength [info level 0]] == 2 } {
	return -code error "wrong # args: should be \"$name swap option ?arg arg ...?\""
    }
    
    # Split the args into command and args components
    set sub __swap_$cmd
    if { [llength [info commands ::struct::matrix::$sub]] == 0 } {
	set optlist [lsort [info commands ::struct::matrix::__swap_*]]
	set xlist {}
	foreach p $optlist {
	    set p [namespace tail $p]
	    lappend xlist [string range $p 7 end]
	}
	set optlist [linsert [join $xlist ", "] "end-1" "or"]
	return -code error \
		"bad option \"$cmd\": must be $optlist"
    }
    uplevel 1 [linsert $args 0 ::struct::matrix::$sub $name]
}

# ::struct::matrix::__add_column --
#
#	Extends the matrix by one column and then acts like
#	"setcolumn" (see below) on this new column if there were
#	"values" supplied. Without "values" the new cells will be set
#	to the empty string. The new column is appended immediately
#	behind the last existing column.
#
# Arguments:
#	name	Name of the matrix object.
#	values	Optional values to set into the new row.
#
# Results:
#	None.

proc ::struct::matrix::__add_column {name {values {}}} {
    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows
    variable ${name}::rowh

    if {[set l [llength $values]] < $rows} {
	# Missing values. Fill up with empty strings

	for {} {$l < $rows} {incr l} {
	    lappend values {}
	}
    } elseif {[llength $values] > $rows} {
	# To many values. Remove the superfluous items
	set values [lrange $values 0 [expr {$rows - 1}]]
    }

    # "values" now contains the information to set into the array.
    # Regarding the width and height caches:

    # - The new column is not added to the width cache, the other
    #   columns are not touched, the cache therefore unchanged.
    # - The rows are either removed from the height cache or left
    #   unchanged, depending on the contents set into the cell.

    set r 0
    foreach v $values {
	if {$v != {}} {
	    # Data changed unpredictably, invalidate cache
	    catch {unset rowh($r)}
	} ; # {else leave the row unchanged}
	set data($columns,$r) $v
	incr r
    }
    incr columns
    return
}

# ::struct::matrix::__add_row --
#
#	Extends the matrix by one row and then acts like "setrow" (see
#	below) on this new row if there were "values"
#	supplied. Without "values" the new cells will be set to the
#	empty string. The new row is appended immediately behind the
#	last existing row.
#
# Arguments:
#	name	Name of the matrix object.
#	values	Optional values to set into the new row.
#
# Results:
#	None.

proc ::struct::matrix::__add_row {name {values {}}} {
    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows
    variable ${name}::colw

    if {[set l [llength $values]] < $columns} {
	# Missing values. Fill up with empty strings

	for {} {$l < $columns} {incr l} {
	    lappend values {}
	}
    } elseif {[llength $values] > $columns} {
	# To many values. Remove the superfluous items
	set values [lrange $values 0 [expr {$columns - 1}]]
    }

    # "values" now contains the information to set into the array.
    # Regarding the width and height caches:

    # - The new row is not added to the height cache, the other
    #   rows are not touched, the cache therefore unchanged.
    # - The columns are either removed from the width cache or left
    #   unchanged, depending on the contents set into the cell.

    set c 0
    foreach v $values {
	if {$v != {}} {
	    # Data changed unpredictably, invalidate cache
	    catch {unset colw($c)}
	} ; # {else leave the row unchanged}
	set data($c,$rows) $v
	incr c
    }
    incr rows
    return
}

# ::struct::matrix::__add_columns --
#
#	Extends the matrix by "n" columns. The new cells will be set
#	to the empty string. The new columns are appended immediately
#	behind the last existing column. A value of "n" equal to or
#	smaller than 0 is not allowed.
#
# Arguments:
#	name	Name of the matrix object.
#	n	The number of new columns to create.
#
# Results:
#	None.

proc ::struct::matrix::__add_columns {name n} {
    if {$n <= 0} {
	return -code error "A value of n <= 0 is not allowed"
    }
    AddColumns $name $n
    return
}

proc ::struct::matrix::AddColumns {name n} {
    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows

    # The new values set into the cell is always the empty
    # string. These have a length and height of 0, i.e. the don't
    # influence cached widths and heights as they are at least that
    # big. IOW there is no need to touch and change the width and
    # height caches.

    while {$n > 0} {
	for {set r 0} {$r < $rows} {incr r} {
	    set data($columns,$r) ""
	}
	incr columns
	incr n -1
    }

    return
}

# ::struct::matrix::__add_rows --
#
#	Extends the matrix by "n" rows. The new cells will be set to
#	the empty string. The new rows are appended immediately behind
#	the last existing row. A value of "n" equal to or smaller than
#	0 is not allowed.
#
# Arguments:
#	name	Name of the matrix object.
#	n	The number of new rows to create.
#
# Results:
#	None.

proc ::struct::matrix::__add_rows {name n} {
    if {$n <= 0} {
	return -code error "A value of n <= 0 is not allowed"
    }
    AddRows $name $n
    return
}

proc ::struct::matrix::AddRows {name n} {
    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows

    # The new values set into the cell is always the empty
    # string. These have a length and height of 0, i.e. the don't
    # influence cached widths and heights as they are at least that
    # big. IOW there is no need to touch and change the width and
    # height caches.

    while {$n > 0} {
	for {set c 0} {$c < $columns} {incr c} {
	    set data($c,$rows) ""
	}
	incr rows
	incr n -1
    }
    return
}

# ::struct::matrix::_cells --
#
#	Returns the number of cells currently managed by the
#	matrix. This is the product of "rows" and "columns".
#
# Arguments:
#	name	Name of the matrix object.
#
# Results:
#	The number of cells in the matrix.

proc ::struct::matrix::_cells {name} {
    variable ${name}::rows
    variable ${name}::columns
    return [expr {$rows * $columns}]
}

# ::struct::matrix::_cellsize --
#
#	Returns the length of the string representation of the value
#	currently contained in the addressed cell.
#
# Arguments:
#	name	Name of the matrix object.
#	column	Column index of the cell to query
#	row	Row index of the cell to query
#
# Results:
#	The number of cells in the matrix.

proc ::struct::matrix::_cellsize {name column row} {
    set column [ChkColumnIndex $name $column]
    set row    [ChkRowIndex    $name $row]

    variable ${name}::data
    return [string length $data($column,$row)]
}

# ::struct::matrix::_columns --
#
#	Returns the number of columns currently managed by the
#	matrix.
#
# Arguments:
#	name	Name of the matrix object.
#
# Results:
#	The number of columns in the matrix.

proc ::struct::matrix::_columns {name} {
    variable ${name}::columns
    return $columns
}

# ::struct::matrix::_columnwidth --
#
#	Returns the length of the longest string representation of all
#	the values currently contained in the cells of the addressed
#	column if these are all spanning only one line. For cell
#	values spanning multiple lines the length of their longest
#	line goes into the computation.
#
# Arguments:
#	name	Name of the matrix object.
#	column	The index of the column whose width is asked for.
#
# Results:
#	See description.

proc ::struct::matrix::_columnwidth {name column} {
    set column [ChkColumnIndex $name $column]

    variable ${name}::colw

    if {![info exists colw($column)]} {
	variable ${name}::rows
	variable ${name}::data

	set width 0
	for {set r 0} {$r < $rows} {incr r} {
	    foreach line [split $data($column,$r) \n] {
		set len [string length $line]
		if {$len > $width} {
		    set width $len
		}
	    }
	}

	set colw($column) $width
    }

    return $colw($column)
}

# ::struct::matrix::__delete_column --
#
#	Deletes the specified column from the matrix and shifts all
#	columns with higher indices one index down.
#
# Arguments:
#	name	Name of the matrix.
#	column	The index of the column to delete.
#
# Results:
#	None.

proc ::struct::matrix::__delete_column {name column} {
    set column [ChkColumnIndex $name $column]

    variable ${name}::data
    variable ${name}::rows
    variable ${name}::columns
    variable ${name}::colw
    variable ${name}::rowh

    # Move all data from the higher columns down and then delete the
    # superfluous data in the old last column. Move the data in the
    # width cache too, take partial fill into account there too.
    # Invalidate the height cache for all rows.

    for {set r 0} {$r < $rows} {incr r} {
	for {set c $column; set cn [expr {$c + 1}]} {$cn < $columns} {incr c ; incr cn} {
	    set data($c,$r) $data($cn,$r)
	    if {[info exists colw($cn)]} {
		set colw($c) $colw($cn)
		unset colw($cn)
	    }
	}
	unset data($c,$r)
	catch {unset rowh($r)}
    }
    incr columns -1
    return
}

# ::struct::matrix::__delete_columns --
#
#	Shrink the matrix by "n" columns (from the right).
#	A value of "n" equal to or smaller than 0 is not
#	allowed, nor is "n" allowed to be greater than the
#	number of columns in the matrix.
#
# Arguments:
#	name	Name of the matrix object.
#	n	The number of columns to remove.
#
# Results:
#	None.

proc ::struct::matrix::__delete_columns {name n} {
    if {$n <= 0} {
	return -code error "A value of n <= 0 is not allowed"
    }

    variable ${name}::columns

    if {$n > $columns} {
	return -code error "A value of n > #columns is not allowed"
    }

    DeleteColumns $name $n
    return
}

# ::struct::matrix::__delete_row --
#
#	Deletes the specified row from the matrix and shifts all
#	row with higher indices one index down.
#
# Arguments:
#	name	Name of the matrix.
#	row	The index of the row to delete.
#
# Results:
#	None.

proc ::struct::matrix::__delete_row {name row} {
    set row [ChkRowIndex $name $row]

    variable ${name}::data
    variable ${name}::rows
    variable ${name}::columns
    variable ${name}::colw
    variable ${name}::rowh

    # Move all data from the higher rows down and then delete the
    # superfluous data in the old last row. Move the data in the
    # height cache too, take partial fill into account there too.
    # Invalidate the width cache for all columns.

    for {set c 0} {$c < $columns} {incr c} {
	for {set r $row; set rn [expr {$r + 1}]} {$rn < $rows} {incr r ; incr rn} {
	    set data($c,$r) $data($c,$rn)
	    if {[info exists rowh($rn)]} {
		set rowh($r) $rowh($rn)
		unset rowh($rn)
	    }
	}
	unset data($c,$r)
	catch {unset colw($c)}
    }
    incr rows -1
    return
}

# ::struct::matrix::__delete_rows --
#
#	Shrink the matrix by "n" rows (from the bottom).
#	A value of "n" equal to or smaller than 0 is not
#	allowed, nor is "n" allowed to be greater than the
#	number of rows in the matrix.
#
# Arguments:
#	name	Name of the matrix object.
#	n	The number of rows to remove.
#
# Results:
#	None.

proc ::struct::matrix::__delete_rows {name n} {
    if {$n <= 0} {
	return -code error "A value of n <= 0 is not allowed"
    }

    variable ${name}::rows

    if {$n > $rows} {
	return -code error "A value of n > #rows is not allowed"
    }

    DeleteRows $name $n
    return
}

# ::struct::matrix::_deserialize --
#
#	Assignment operator. Copies a serialization into the
#       destination, destroying the original information.
#
# Arguments:
#	name	Name of the matrix object we are copying into.
#	serial	Serialized matrix to copy from.
#
# Results:
#	Nothing.

proc ::struct::matrix::_deserialize {name serial} {
    # As we destroy the original matrix as part of
    # the copying process we don't have to deal
    # with issues like node names from the new matrix
    # interfering with the old ...

    # I. Get the serialization of the source matrix
    #    and check it for validity.

    CheckSerialization $serial r c values

    # Get all the relevant data into the scope

    variable ${name}::rows
    variable ${name}::columns

    # Resize the destination matrix for the new data

    if {$r > $rows} {
	AddRows $name [expr {$r - $rows}]
    } elseif {$r < $rows} {
	DeleteRows $name [expr {$rows - $r}]
    }
    if {$c > $columns} {
	AddColumns $name [expr {$c - $columns}]
    } elseif {$c < $columns} {
	DeleteColumns $name [expr {$columns - $c}]
    }

    set rows    $r
    set columns $c

    # Copy the new data over the old information.

    set row 0
    foreach rv $values {
	SetRow $name $row $rv
	incr row
    }
    while {$row < $rows} {
	# Fill with empty rows if there are not enough.
	SetRow $name $row {}
	incr row
    }
    return
}

# ::struct::matrix::_destroy --
#
#	Destroy a matrix, including its associated command and data storage.
#
# Arguments:
#	name	Name of the matrix to destroy.
#
# Results:
#	None.

proc ::struct::matrix::_destroy {name} {
    variable ${name}::link

    # Unlink all existing arrays before destroying the object so that
    # we don't leave dangling references / traces.

    foreach avar [array names link] {
	_unlink $name $avar
    }

    namespace delete $name
    interp alias {}  $name {}
}

# ::struct::matrix::__format_2string --
#
#	Formats the matrix using the specified report object and
#	returns the string containing the result of this
#	operation. The report has to support the "printmatrix" method.
#
# Arguments:
#	name	Name of the matrix.
#	report	Name of the report object specifying the formatting.
#
# Results:
#	A string containing the formatting result.

proc ::struct::matrix::__format_2string {name {report {}}} {
    if {$report == {}} {
	# Use an internal hardwired simple report to format the matrix.
	# 1. Go through all columns and compute the column widths.
	# 2. Then iterate through all rows and dump then into a
	#    string, formatted to the number of characters per columns

	array set cw {}
	set cols [_columns $name]
	for {set c 0} {$c < $cols} {incr c} {
	    set cw($c) [_columnwidth $name $c]
	}

	set result [list]
	set n [_rows $name]
	for {set r 0} {$r < $n} {incr r} {
	    set rh [_rowheight $name $r]
	    if {$rh < 2} {
		# Simple row.
		set line [list]
		for {set c 0} {$c < $cols} {incr c} {
		    set val [__get_cell $name $c $r]
		    lappend line "$val[string repeat " " [expr {$cw($c)-[string length $val]}]]"
		}
		lappend result [join $line " "]
	    } else {
		# Complex row, multiple passes
		for {set h 0} {$h < $rh} {incr h} {
		    set line [list]
		    for {set c 0} {$c < $cols} {incr c} {
			set val [lindex [split [__get_cell $name $c $r] \n] $h]
			lappend line "$val[string repeat " " [expr {$cw($c)-[string length $val]}]]"
		    }
		    lappend result [join $line " "]
		}
	    }
	}
	return [join $result \n]
    } else {
	return [$report printmatrix $name]
    }
}

# ::struct::matrix::__format_2chan --
#
#	Formats the matrix using the specified report object and
#	writes the string containing the result of this operation into
#	the channel. The report has to support the
#	"printmatrix2channel" method.
#
# Arguments:
#	name	Name of the matrix.
#	report	Name of the report object specifying the formatting.
#	chan	Handle of the channel to write to.
#
# Results:
#	None.

proc ::struct::matrix::__format_2chan {name {report {}} {chan stdout}} {
    if {$report == {}} {
	# Use an internal hardwired simple report to format the matrix.
	# We delegate this to the string formatter and print its result.
	puts -nonewline [__format_2string $name]
    } else {
	$report printmatrix2channel $name $chan
    }
    return
}

# ::struct::matrix::__get_cell --
#
#	Returns the value currently contained in the cell identified
#	by row and column index.
#
# Arguments:
#	name	Name of the matrix.
#	column	Column index of the addressed cell.
#	row	Row index of the addressed cell.
#
# Results:
#	value	Value currently stored in the addressed cell.

proc ::struct::matrix::__get_cell {name column row} {
    set column [ChkColumnIndex $name $column]
    set row    [ChkRowIndex    $name $row]

    variable ${name}::data
    return $data($column,$row)
}

# ::struct::matrix::__get_column --
#
#	Returns a list containing the values from all cells in the
#	column identified by the index. The contents of the cell in
#	row 0 are stored as the first element of this list.
#
# Arguments:
#	name	Name of the matrix.
#	column	Column index of the addressed cell.
#
# Results:
#	List of values stored in the addressed row.

proc ::struct::matrix::__get_column {name column} {
    set column [ChkColumnIndex $name $column]
    return     [GetColumn      $name $column]
}

proc ::struct::matrix::GetColumn {name column} {
    variable ${name}::data
    variable ${name}::rows

    set result [list]
    for {set r 0} {$r < $rows} {incr r} {
	lappend result $data($column,$r)
    }
    return $result
}

# ::struct::matrix::__get_rect --
#
#	Returns a list of lists of cell values. The values stored in
#	the result come from the submatrix whose top-left and
#	bottom-right cells are specified by "column_tl", "row_tl" and
#	"column_br", "row_br" resp. Note that the following equations
#	have to be true: column_tl <= column_br and row_tl <= row_br.
#	The result is organized as follows: The outer list is the list
#	of rows, its elements are lists representing a single row. The
#	row with the smallest index is the first element of the outer
#	list. The elements of the row lists represent the selected
#	cell values. The cell with the smallest index is the first
#	element in each row list.
#
# Arguments:
#	name		Name of the matrix.
#	column_tl	Column index of the top-left cell of the area.
#	row_tl		Row index of the top-left cell of the the area
#	column_br	Column index of the bottom-right cell of the area.
#	row_br		Row index of the bottom-right cell of the the area
#
# Results:
#	List of a list of values stored in the addressed area.

proc ::struct::matrix::__get_rect {name column_tl row_tl column_br row_br} {
    set column_tl [ChkColumnIndex $name $column_tl]
    set row_tl    [ChkRowIndex    $name $row_tl]
    set column_br [ChkColumnIndex $name $column_br]
    set row_br    [ChkRowIndex    $name $row_br]

    if {
	($column_tl > $column_br) ||
	($row_tl    > $row_br)
    } {
	return -code error "Invalid cell indices, wrong ordering"
    }
    return [GetRect $name $column_tl $row_tl $column_br $row_br]
}

proc ::struct::matrix::GetRect {name column_tl row_tl column_br row_br} {
    variable ${name}::data
    set result [list]

    for {set r $row_tl} {$r <= $row_br} {incr r} {
	set row [list]
	for {set c $column_tl} {$c <= $column_br} {incr c} {
	    lappend row $data($c,$r)
	}
	lappend result $row
    }

    return $result
}

# ::struct::matrix::__get_row --
#
#	Returns a list containing the values from all cells in the
#	row identified by the index. The contents of the cell in
#	column 0 are stored as the first element of this list.
#
# Arguments:
#	name	Name of the matrix.
#	row	Row index of the addressed cell.
#
# Results:
#	List of values stored in the addressed row.

proc ::struct::matrix::__get_row {name row} {
    set row [ChkRowIndex $name $row]
    return  [GetRow      $name $row]
}

proc ::struct::matrix::GetRow {name row} {
    variable ${name}::data
    variable ${name}::columns

    set result [list]
    for {set c 0} {$c < $columns} {incr c} {
	lappend result $data($c,$row)
    }
    return $result
}

# ::struct::matrix::__insert_column --
#
#	Extends the matrix by one column and then acts like
#	"setcolumn" (see below) on this new column if there were
#	"values" supplied. Without "values" the new cells will be set
#	to the empty string. The new column is inserted just before
#	the column specified by the given index. This means, if
#	"column" is less than or equal to zero, then the new column is
#	inserted at the beginning of the matrix, before the first
#	column. If "column" has the value "Bend", or if it is greater
#	than or equal to the number of columns in the matrix, then the
#	new column is appended to the matrix, behind the last
#	column. The old column at the chosen index and all columns
#	with higher indices are shifted one index upward.
#
# Arguments:
#	name	Name of the matrix.
#	column	Index of the column where to insert.
#	values	Optional values to set the cells to.
#
# Results:
#	None.

proc ::struct::matrix::__insert_column {name column {values {}}} {
    # Allow both negative and too big indices.
    set column [ChkColumnIndexAll $name $column]

    variable ${name}::columns

    if {$column > $columns} {
	# Same as 'addcolumn'
	__add_column $name $values
	return
    }

    variable ${name}::data
    variable ${name}::rows
    variable ${name}::rowh
    variable ${name}::colw

    set firstcol $column
    if {$firstcol < 0} {
	set firstcol 0
    }

    if {[set l [llength $values]] < $rows} {
	# Missing values. Fill up with empty strings

	for {} {$l < $rows} {incr l} {
	    lappend values {}
	}
    } elseif {[llength $values] > $rows} {
	# To many values. Remove the superfluous items
	set values [lrange $values 0 [expr {$rows - 1}]]
    }

    # "values" now contains the information to set into the array.
    # Regarding the width and height caches:
    # Invalidate all rows, move all columns

    # Move all data from the higher columns one up and then insert the
    # new data into the freed space. Move the data in the
    # width cache too, take partial fill into account there too.
    # Invalidate the height cache for all rows.

    for {set r 0} {$r < $rows} {incr r} {
	for {set cn $columns ; set c [expr {$cn - 1}]} {$c >= $firstcol} {incr c -1 ; incr cn -1} {
	    set data($cn,$r) $data($c,$r)
	    if {[info exists colw($c)]} {
		set colw($cn) $colw($c)
		unset colw($c)
	    }
	}
	set data($firstcol,$r) [lindex $values $r]
	catch {unset rowh($r)}
    }
    incr columns
    return
}

# ::struct::matrix::__insert_row --
#
#	Extends the matrix by one row and then acts like "setrow" (see
#	below) on this new row if there were "values"
#	supplied. Without "values" the new cells will be set to the
#	empty string. The new row is inserted just before the row
#	specified by the given index. This means, if "row" is less
#	than or equal to zero, then the new row is inserted at the
#	beginning of the matrix, before the first row. If "row" has
#	the value "end", or if it is greater than or equal to the
#	number of rows in the matrix, then the new row is appended to
#	the matrix, behind the last row. The old row at that index and
#	all rows with higher indices are shifted one index upward.
#
# Arguments:
#	name	Name of the matrix.
#	row	Index of the row where to insert.
#	values	Optional values to set the cells to.
#
# Results:
#	None.

proc ::struct::matrix::__insert_row {name row {values {}}} {
    # Allow both negative and too big indices.
    set row [ChkRowIndexAll $name $row]

    variable ${name}::rows

    if {$row > $rows} {
	# Same as 'addrow'
	__add_row $name $values
	return
    }

    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rowh
    variable ${name}::colw

    set firstrow $row
    if {$firstrow < 0} {
	set firstrow 0
    }

    if {[set l [llength $values]] < $columns} {
	# Missing values. Fill up with empty strings

	for {} {$l < $columns} {incr l} {
	    lappend values {}
	}
    } elseif {[llength $values] > $columns} {
	# To many values. Remove the superfluous items
	set values [lrange $values 0 [expr {$columns - 1}]]
    }

    # "values" now contains the information to set into the array.
    # Regarding the width and height caches:
    # Invalidate all columns, move all rows

    # Move all data from the higher rows one up and then insert the
    # new data into the freed space. Move the data in the
    # height cache too, take partial fill into account there too.
    # Invalidate the width cache for all columns.

    for {set c 0} {$c < $columns} {incr c} {
	for {set rn $rows ; set r [expr {$rn - 1}]} {$r >= $firstrow} {incr r -1 ; incr rn -1} {
	    set data($c,$rn) $data($c,$r)
	    if {[info exists rowh($r)]} {
		set rowh($rn) $rowh($r)
		unset rowh($r)
	    }
	}
	set data($c,$firstrow) [lindex $values $c]
	catch {unset colw($c)}
    }
    incr rows
    return
}

# ::struct::matrix::_link --
#
#	Links the matrix to the specified array variable. This means
#	that the contents of all cells in the matrix is stored in the
#	array too, with all changes to the matrix propagated there
#	too. The contents of the cell "(column,row)" is stored in the
#	array using the key "column,row". If the option "-transpose"
#	is specified the key "row,column" will be used instead. It is
#	possible to link the matrix to more than one array. Note that
#	the link is bidirectional, i.e. changes to the array are
#	mirrored in the matrix too.
#
# Arguments:
#	name	Name of the matrix object.
#	option	Either empty of '-transpose'.
#	avar	Name of the variable to link to
#
# Results:
#	None

proc ::struct::matrix::_link {name args} {
    switch -exact -- [llength $args] {
	0 {
	    return -code error "$name: wrong # args: link ?-transpose? arrayvariable"
	}
	1 {
	    set transpose 0
	    set variable  [lindex $args 0]
	}
	2 {
	    foreach {t variable} $args break
	    if {[string compare $t -transpose]} {
		return -code error "$name: illegal syntax: link ?-transpose? arrayvariable"
	    }
	    set transpose 1
	}
	default {
	    return -code error "$name: wrong # args: link ?-transpose? arrayvariable"
	}
    }

    variable ${name}::link

    if {[info exists link($variable)]} {
	return -code error "$name link: Variable \"$variable\" already linked to matrix"
    }

    # Ok, a new variable we are linked to. Record this information,
    # dump our current contents into the array, at last generate the
    # traces actually performing the link.

    set link($variable) $transpose

    upvar #0 $variable array
    variable ${name}::data

    foreach key [array names data] {
	foreach {c r} [split $key ,] break
	if {$transpose} {
	    set array($r,$c) $data($key)
	} else {
	    set array($c,$r) $data($key)
	}
    }

    trace variable array wu [list ::struct::matrix::MatTraceIn  $variable $name]
    trace variable data  w  [list ::struct::matrix::MatTraceOut $variable $name]
    return
}

# ::struct::matrix::_links --
#
#	Retrieves the names of all array variable the matrix is
#	officially linked to.
#
# Arguments:
#	name	Name of the matrix object.
#
# Results:
#	List of variables the matrix is linked to.

proc ::struct::matrix::_links {name} {
    variable ${name}::link
    return [array names link]
}

# ::struct::matrix::_rowheight --
#
#	Returns the height of the specified row in lines. This is the
#	highest number of lines spanned by a cell over all cells in
#	the row.
#
# Arguments:
#	name	Name of the matrix
#	row	Index of the row queried for its height
#
# Results:
#	The height of the specified row in lines.

proc ::struct::matrix::_rowheight {name row} {
    set row [ChkRowIndex $name $row]

    variable ${name}::rowh

    if {![info exists rowh($row)]} {
	variable ${name}::columns
	variable ${name}::data

	set height 1
	for {set c 0} {$c < $columns} {incr c} {
	    set cheight [llength [split $data($c,$row) \n]]
	    if {$cheight > $height} {
		set height $cheight
	    }
	}

	set rowh($row) $height
    }
    return $rowh($row)
}

# ::struct::matrix::_rows --
#
#	Returns the number of rows currently managed by the matrix.
#
# Arguments:
#	name	Name of the matrix object.
#
# Results:
#	The number of rows in the matrix.

proc ::struct::matrix::_rows {name} {
    variable ${name}::rows
    return $rows
}

# ::struct::matrix::_serialize --
#
#	Serialize a matrix object (partially) into a transportable value.
#	If only a rectangle is serialized the result will be a sub-
#	matrix in the mathematical sense of the word.
#
# Arguments:
#	name	Name of the matrix.
#	args	rectangle to place into the serialized matrix
#
# Results:
#	A list structure describing the part of the matrix which was serialized.

proc ::struct::matrix::_serialize {name args} {

    # all - boolean flag - set if and only if the all nodes of the
    # matrix are chosen for serialization. Because if that is true we
    # can skip the step finding the relevant arcs and simply take all
    # arcs.

    set nargs [llength $args]
    if {($nargs != 0) && ($nargs != 4)} {
	return -code error "$name: wrong # args: serialize ?column_tl row_tl column_br row_br?"
    }

    variable ${name}::rows
    variable ${name}::columns

    if {$nargs == 4} {
	foreach {column_tl row_tl column_br row_br} $args break

	set column_tl [ChkColumnIndex $name $column_tl]
	set row_tl    [ChkRowIndex    $name $row_tl]
	set column_br [ChkColumnIndex $name $column_br]
	set row_br    [ChkRowIndex    $name $row_br]

	if {
	    ($column_tl > $column_br) ||
	    ($row_tl    > $row_br)
	} {
	    return -code error "Invalid cell indices, wrong ordering"
	}
	set rn [expr {$row_br    - $row_tl    + 1}]
	set cn [expr {$column_br - $column_tl + 1}]
    } else {
	set column_tl 0
	set row_tl    0
	set column_br [expr {$columns - 1}]
	set row_br    [expr {$rows - 1}]
	set rn $rows
	set cn $columns
    }

    # We could optimize and remove empty cells to the right and rows
    # to the bottom. For now we don't.

    return [list \
	    $rn $cn \
	    [GetRect $name $column_tl $row_tl $column_br $row_br]]
}

# ::struct::matrix::__set_cell --
#
#	Sets the value in the cell identified by row and column index
#	to the data in the third argument.
#
# Arguments:
#	name	Name of the matrix object.
#	column	Column index of the cell to set.
#	row	Row index of the cell to set.
#	value	The new value of the cell.
#
# Results:
#	None.
 
proc ::struct::matrix::__set_cell {name column row value} {
    set column [ChkColumnIndex $name $column]
    set row    [ChkRowIndex    $name $row]

    variable ${name}::data

    if {![string compare $value $data($column,$row)]} {
	# No change, ignore call!
	return
    }

    set data($column,$row) $value

    if {$value != {}} {
	variable ${name}::colw
	variable ${name}::rowh
	catch {unset colw($column)}
	catch {unset rowh($row)}
    }
    return
}

# ::struct::matrix::__set_column --
#
#	Sets the values in the cells identified by the column index to
#	the elements of the list provided as the third argument. Each
#	element of the list is assigned to one cell, with the first
#	element going into the cell in row 0 and then upward. If there
#	are less values in the list than there are rows the remaining
#	rows are set to the empty string. If there are more values in
#	the list than there are rows the superfluous elements are
#	ignored. The matrix is not extended by this operation.
#
# Arguments:
#	name	Name of the matrix.
#	column	Index of the column to set.
#	values	Values to set into the column.
#
# Results:
#	None.

proc ::struct::matrix::__set_column {name column values} {
    set column [ChkColumnIndex $name $column]

    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows
    variable ${name}::rowh
    variable ${name}::colw

    if {[set l [llength $values]] < $rows} {
	# Missing values. Fill up with empty strings

	for {} {$l < $rows} {incr l} {
	    lappend values {}
	}
    } elseif {[llength $values] > $rows} {
	# To many values. Remove the superfluous items
	set values [lrange $values 0 [expr {$rows - 1}]]
    }

    # "values" now contains the information to set into the array.
    # Regarding the width and height caches:

    # - Invalidate the column in the width cache.
    # - The rows are either removed from the height cache or left
    #   unchanged, depending on the contents set into the cell.

    set r 0
    foreach v $values {
	if {$v != {}} {
	    # Data changed unpredictably, invalidate cache
	    catch {unset rowh($r)}
	} ; # {else leave the row unchanged}
	set data($column,$r) $v
	incr r
    }
    catch {unset colw($column)}
    return
}

# ::struct::matrix::__set_rect --
#
#	Takes a list of lists of cell values and writes them into the
#	submatrix whose top-left cell is specified by the two
#	indices. If the sublists of the outer list are not of equal
#	length the shorter sublists will be filled with empty strings
#	to the length of the longest sublist. If the submatrix
#	specified by the top-left cell and the number of rows and
#	columns in the "values" extends beyond the matrix we are
#	modifying the over-extending parts of the values are ignored,
#	i.e. essentially cut off. This subcommand expects its input in
#	the format as returned by "getrect".
#
# Arguments:
#	name	Name of the matrix object.
#	column	Column index of the topleft cell to set.
#	row	Row index of the topleft cell to set.
#	values	Values to set.
#
# Results:
#	None.

proc ::struct::matrix::__set_rect {name column row values} {
    # Allow negative indices!
    set column [ChkColumnIndexNeg $name $column]
    set row    [ChkRowIndexNeg    $name $row]

    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows
    variable ${name}::colw
    variable ${name}::rowh

    if {$row < 0} {
	# Remove rows from the head of values to restrict it to the
	# overlapping area.

	set values [lrange $values [expr {0 - $row}] end]
	set row 0
    }

    # Restrict it at the end too.
    if {($row + [llength $values]) > $rows} {
	set values [lrange $values 0 [expr {$rows - $row - 1}]]
    }

    # Same for columns, but store it in some vars as this is required
    # in a loop.
    set firstcol 0
    if {$column < 0} {
	set firstcol [expr {0 - $column}]
	set column 0
    }

    # Now pan through values and area and copy the external data into
    # the matrix.

    set r $row
    foreach line $values {
	set line [lrange $line $firstcol end]

	set l [expr {$column + [llength $line]}]
	if {$l > $columns} {
	    set line [lrange $line 0 [expr {$columns - $column - 1}]]
	} elseif {$l < [expr {$columns - $firstcol}]} {
	    # We have to take the offset into the line into account
	    # or we add fillers we don't need, overwriting part of the
	    # data array we shouldn't.

	    for {} {$l < [expr {$columns - $firstcol}]} {incr l} {
		lappend line {}
	    }
	}

	set c $column
	foreach cell $line {
	    if {$cell != {}} {
		catch {unset rowh($r)}
		catch {unset colw($c)}
	    }
	    set data($c,$r) $cell
	    incr c
	}
	incr r
    }
    return
}

# ::struct::matrix::__set_row --
#
#	Sets the values in the cells identified by the row index to
#	the elements of the list provided as the third argument. Each
#	element of the list is assigned to one cell, with the first
#	element going into the cell in column 0 and then upward. If
#	there are less values in the list than there are columns the
#	remaining columns are set to the empty string. If there are
#	more values in the list than there are columns the superfluous
#	elements are ignored. The matrix is not extended by this
#	operation.
#
# Arguments:
#	name	Name of the matrix.
#	row	Index of the row to set.
#	values	Values to set into the row.
#
# Results:
#	None.

proc ::struct::matrix::__set_row {name row values} {
    set row [ChkRowIndex $name $row]
    SetRow $name $row $values
}

proc ::struct::matrix::SetRow {name row values} {
    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rows
    variable ${name}::colw
    variable ${name}::rowh

    if {[set l [llength $values]] < $columns} {
	# Missing values. Fill up with empty strings

	for {} {$l < $columns} {incr l} {
	    lappend values {}
	}
    } elseif {[llength $values] > $columns} {
	# To many values. Remove the superfluous items
	set values [lrange $values 0 [expr {$columns - 1}]]
    }

    # "values" now contains the information to set into the array.
    # Regarding the width and height caches:

    # - Invalidate the row in the height cache.
    # - The columns are either removed from the width cache or left
    #   unchanged, depending on the contents set into the cell.

    set c 0
    foreach v $values {
	if {$v != {}} {
	    # Data changed unpredictably, invalidate cache
	    catch {unset colw($c)}
	} ; # {else leave the row unchanged}
	set data($c,$row) $v
	incr c
    }
    catch {unset rowh($row)}
    return
}

# ::struct::matrix::__swap_columns --
#
#	Swaps the contents of the two specified columns.
#
# Arguments:
#	name		Name of the matrix.
#	column_a	Index of the first column to swap
#	column_b	Index of the second column to swap
#
# Results:
#	None.

proc ::struct::matrix::__swap_columns {name column_a column_b} {
    set column_a [ChkColumnIndex $name $column_a]
    set column_b [ChkColumnIndex $name $column_b]
    return [SwapColumns $name $column_a $column_b]
}

proc ::struct::matrix::SwapColumns {name column_a column_b} {
    variable ${name}::data
    variable ${name}::rows
    variable ${name}::colw

    # Note: This operation does not influence the height cache for all
    # rows and the width cache only insofar as its contents has to be
    # swapped too for the two columns we are touching. Note that the
    # cache might be partially filled or not at all, so we don't have
    # to "swap" in some situations.

    for {set r 0} {$r < $rows} {incr r} {
	set tmp                $data($column_a,$r)
	set data($column_a,$r) $data($column_b,$r)
	set data($column_b,$r) $tmp
    }

    set cwa [info exists colw($column_a)]
    set cwb [info exists colw($column_b)]

    if {$cwa && $cwb} {
	set tmp             $colw($column_a)
	set colw($column_a) $colw($column_b)
	set colw($column_b) $tmp
    } elseif {$cwa} {
	# Move contents, don't swap.
	set   colw($column_b) $colw($column_a)
	unset colw($column_a)
    } elseif {$cwb} {
	# Move contents, don't swap.
	set   colw($column_a) $colw($column_b)
	unset colw($column_b)
    } ; # else {nothing to do at all}
    return
}

# ::struct::matrix::__swap_rows --
#
#	Swaps the contents of the two specified rows.
#
# Arguments:
#	name	Name of the matrix.
#	row_a	Index of the first row to swap
#	row_b	Index of the second row to swap
#
# Results:
#	None.

proc ::struct::matrix::__swap_rows {name row_a row_b} {
    set row_a [ChkRowIndex $name $row_a]
    set row_b [ChkRowIndex $name $row_b]
    return [SwapRows $name $row_a $row_b]
}

proc ::struct::matrix::SwapRows {name row_a row_b} {
    variable ${name}::data
    variable ${name}::columns
    variable ${name}::rowh

    # Note: This operation does not influence the width cache for all
    # columns and the height cache only insofar as its contents has to be
    # swapped too for the two rows we are touching. Note that the
    # cache might be partially filled or not at all, so we don't have
    # to "swap" in some situations.

    for {set c 0} {$c < $columns} {incr c} {
	set tmp             $data($c,$row_a)
	set data($c,$row_a) $data($c,$row_b)
	set data($c,$row_b) $tmp
    }

    set rha [info exists rowh($row_a)]
    set rhb [info exists rowh($row_b)]

    if {$rha && $rhb} {
	set tmp          $rowh($row_a)
	set rowh($row_a) $rowh($row_b)
	set rowh($row_b) $tmp
    } elseif {$rha} {
	# Move contents, don't swap.
	set   rowh($row_b) $rowh($row_a)
	unset rowh($row_a)
    } elseif {$rhb} {
	# Move contents, don't swap.
	set   rowh($row_a) $rowh($row_b)
	unset rowh($row_b)
    } ; # else {nothing to do at all}
    return
}

# ::struct::matrix::_transpose --
#
#	Exchanges rows and columns of the matrix
#
# Arguments:
#	name		Name of the matrix.
#
# Results:
#	None.

proc ::struct::matrix::_transpose {name} {
    variable ${name}::rows
    variable ${name}::columns

    if {$rows == 0} {
	# Change the dimensions. 
	# There is no data to shift.
	# The row/col caches are empty too.

	set rows    $columns 
	set columns 0
	return

    } elseif {$columns == 0} {
	# Change the dimensions. 
	# There is no data to shift.
	# The row/col caches are empty too.

	set columns $rows
	set rows    0
	return
    }

    variable ${name}::data
    variable ${name}::rowh
    variable ${name}::colw

    # Exchanging the row/col caches is easy, independent of the actual
    # dimensions of the matrix.

    set rhc [array get rowh]
    set cwc [array get colw]

    unset rowh ; array set rowh $cwc
    unset colw ; array set colw $rhc

    if {$rows == $columns} {
	# A square matrix. We have to swap data around, but there is
	# need to resize any of the arrays. Only the core is present.

	set n $columns

    } elseif {$rows > $columns} {
	# Rectangular matrix, we have to delete rows, and add columns.

	for {set r $columns} {$r < $rows} {incr r} {
	    for {set c 0} {$c < $columns} {incr c} {
		set   data($r,$c) $data($c,$r)
		unset data($c,$r)
	    }
	}

	set n $columns ; # Size of the core.
    } else {
	# rows < columns. Rectangular matrix, we have to delete
	# columns, and add rows.

	for {set c $rows} {$c < $columns} {incr c} {
	    for {set r 0} {$r < $rows} {incr r} {
		set   data($r,$c) $data($c,$r)
		unset data($c,$r)
	    }
	}

	set n $rows ; # Size of the core.
    }

    set tmp     $rows
    set rows    $columns
    set columns $tmp

    # Whatever the actual dimensions, a square core is always
    # present. The data of this core is now shuffled

    for {set i 0} {$i < $n} {incr i} {
	for {set j $i ; incr j} {$j < $n} {incr j} {
	    set tmp $data($i,$j)
	    set data($i,$j) $data($j,$i)
	    set data($j,$i) $tmp
	}
    }
    return
}

# ::struct::matrix::_unlink --
#
#	Removes the link between the matrix and the specified
#	arrayvariable, if there is one.
#
# Arguments:
#	name	Name of the matrix.
#	avar	Name of the linked array.
#
# Results:
#	None.

proc ::struct::matrix::_unlink {name avar} {

    variable ${name}::link

    if {![info exists link($avar)]} {
	# Ignore unlinking of unknown variables.
	return
    }

    # Delete the traces first, then remove the link management
    # information from the object.

    upvar #0 $avar    array
    variable ${name}::data

    trace vdelete array wu [list ::struct::matrix::MatTraceIn  $avar $name]
    trace vdelete date  w  [list ::struct::matrix::MatTraceOut $avar $name]

    unset link($avar)
    return
}

# ::struct::matrix::ChkColumnIndex --
#
#	Helper to check and transform column indices. Returns the
#	absolute index number belonging to the specified
#	index. Rejects indices out of the valid range of columns.
#
# Arguments:
#	matrix	Matrix to look at
#	column	The incoming index to check and transform
#
# Results:
#	The absolute index to the column

proc ::struct::matrix::ChkColumnIndex {name column} {
    variable ${name}::columns

    switch -regexp -- $column {
	{end-[0-9]+} {
	    set column [string map {end- ""} $column]
	    set cc [expr {$columns - 1 - $column}]
	    if {($cc < 0) || ($cc >= $columns)} {
		return -code error "bad column index end-$column, column does not exist"
	    }
	    return $cc
	}
	end {
	    if {$columns <= 0} {
		return -code error "bad column index $column, column does not exist"
	    }
	    return [expr {$columns - 1}]
	}
	{[0-9]+} {
	    if {($column < 0) || ($column >= $columns)} {
		return -code error "bad column index $column, column does not exist"
	    }
	    return $column
	}
	default {
	    return -code error "bad column index \"$column\", syntax error"
	}
    }
    # Will not come to this place
}

# ::struct::matrix::ChkRowIndex --
#
#	Helper to check and transform row indices. Returns the
#	absolute index number belonging to the specified
#	index. Rejects indices out of the valid range of rows.
#
# Arguments:
#	matrix	Matrix to look at
#	row	The incoming index to check and transform
#
# Results:
#	The absolute index to the row

proc ::struct::matrix::ChkRowIndex {name row} {
    variable ${name}::rows

    switch -regexp -- $row {
	{end-[0-9]+} {
	    set row [string map {end- ""} $row]
	    set rr [expr {$rows - 1 - $row}]
	    if {($rr < 0) || ($rr >= $rows)} {
		return -code error "bad row index end-$row, row does not exist"
	    }
	    return $rr
	}
	end {
	    if {$rows <= 0} {
		return -code error "bad row index $row, row does not exist"
	    }
	    return [expr {$rows - 1}]
	}
	{[0-9]+} {
	    if {($row < 0) || ($row >= $rows)} {
		return -code error "bad row index $row, row does not exist"
	    }
	    return $row
	}
	default {
	    return -code error "bad row index \"$row\", syntax error"
	}
    }
    # Will not come to this place
}

# ::struct::matrix::ChkColumnIndexNeg --
#
#	Helper to check and transform column indices. Returns the
#	absolute index number belonging to the specified
#	index. Rejects indices out of the valid range of columns
#	(Accepts negative indices).
#
# Arguments:
#	matrix	Matrix to look at
#	column	The incoming index to check and transform
#
# Results:
#	The absolute index to the column

proc ::struct::matrix::ChkColumnIndexNeg {name column} {
    variable ${name}::columns

    switch -regexp -- $column {
	{end-[0-9]+} {
	    set column [string map {end- ""} $column]
	    set cc [expr {$columns - 1 - $column}]
	    if {$cc >= $columns} {
		return -code error "bad column index end-$column, column does not exist"
	    }
	    return $cc
	}
	end {
	    return [expr {$columns - 1}]
	}
	{[0-9]+} {
	    if {$column >= $columns} {
		return -code error "bad column index $column, column does not exist"
	    }
	    return $column
	}
	default {
	    return -code error "bad column index \"$column\", syntax error"
	}
    }
    # Will not come to this place
}

# ::struct::matrix::ChkRowIndexNeg --
#
#	Helper to check and transform row indices. Returns the
#	absolute index number belonging to the specified
#	index. Rejects indices out of the valid range of rows
#	(Accepts negative indices).
#
# Arguments:
#	matrix	Matrix to look at
#	row	The incoming index to check and transform
#
# Results:
#	The absolute index to the row

proc ::struct::matrix::ChkRowIndexNeg {name row} {
    variable ${name}::rows

    switch -regexp -- $row {
	{end-[0-9]+} {
	    set row [string map {end- ""} $row]
	    set rr [expr {$rows - 1 - $row}]
	    if {$rr >= $rows} {
		return -code error "bad row index end-$row, row does not exist"
	    }
	    return $rr
	}
	end {
	    return [expr {$rows - 1}]
	}
	{[0-9]+} {
	    if {$row >= $rows} {
		return -code error "bad row index $row, row does not exist"
	    }
	    return $row
	}
	default {
	    return -code error "bad row index \"$row\", syntax error"
	}
    }
    # Will not come to this place
}

# ::struct::matrix::ChkColumnIndexAll --
#
#	Helper to transform column indices. Returns the
#	absolute index number belonging to the specified
#	index.
#
# Arguments:
#	matrix	Matrix to look at
#	column	The incoming index to check and transform
#
# Results:
#	The absolute index to the column

proc ::struct::matrix::ChkColumnIndexAll {name column} {
    variable ${name}::columns

    switch -regexp -- $column {
	{end-[0-9]+} {
	    set column [string map {end- ""} $column]
	    set cc [expr {$columns - 1 - $column}]
	    return $cc
	}
	end {
	    return $columns
	}
	{[0-9]+} {
	    return $column
	}
	default {
	    return -code error "bad column index \"$column\", syntax error"
	}
    }
    # Will not come to this place
}

# ::struct::matrix::ChkRowIndexAll --
#
#	Helper to transform row indices. Returns the
#	absolute index number belonging to the specified
#	index.
#
# Arguments:
#	matrix	Matrix to look at
#	row	The incoming index to check and transform
#
# Results:
#	The absolute index to the row

proc ::struct::matrix::ChkRowIndexAll {name row} {
    variable ${name}::rows

    switch -regexp -- $row {
	{end-[0-9]+} {
	    set row [string map {end- ""} $row]
	    set rr [expr {$rows - 1 - $row}]
	    return $rr
	}
	end {
	    return $rows
	}
	{[0-9]+} {
	    return $row
	}
	default {
	    return -code error "bad row index \"$row\", syntax error"
	}
    }
    # Will not come to this place
}

# ::struct::matrix::MatTraceIn --
#
#	Helper propagating changes made to an array
#	into the matrix the array is linked to.
#
# Arguments:
#	avar		Name of the array which was changed.
#	name		Matrix to write the changes to.
#	var,idx,op	Standard trace arguments
#
# Results:
#	None.

proc ::struct::matrix::MatTraceIn {avar name var idx op} {
    # Propagate changes in the linked array back into the matrix.

    variable ${name}::lock
    if {$lock} {return}

    # We have to cover two possibilities when encountering an "unset" operation ...
    # 1. The external array was destroyed: perform automatic unlink.
    # 2. An individual element was unset:  Set the corresponding cell to the empty string.
    #    See SF Tcllib Bug #532791.

    if {(![string compare $op u]) && ($idx == {})} {
	# Possibility 1: Array was destroyed
	$name unlink $avar
	return
    }

    upvar #0 $avar    array
    variable ${name}::data
    variable ${name}::link

    set transpose $link($avar)
    if {$transpose} {
	foreach {r c} [split $idx ,] break
    } else {
	foreach {c r} [split $idx ,] break
    }

    # Use standard method to propagate the change.
    # => Get automatically index checks, cache updates, ...

    if {![string compare $op u]} {
	# Unset possibility 2: Element was unset.
	# Note: Setting the cell to the empty string will
	# invoke MatTraceOut for this array and thus try
	# to recreate the destroyed element of the array.
	# We don't want this. But we do want to propagate
	# the change to other arrays, as "unset". To do
	# all of this we use another state variable to
	# signal this situation.

	variable ${name}::unset
	set unset $avar

	$name set cell $c $r ""

	set unset {}
	return
    }

    $name set cell $c $r $array($idx)
    return
}

# ::struct::matrix::MatTraceOut --
#
#	Helper propagating changes made to the matrix into the linked arrays.
#
# Arguments:
#	avar		Name of the array to write the changes to.
#	name		Matrix which was changed.
#	var,idx,op	Standard trace arguments
#
# Results:
#	None.

proc ::struct::matrix::MatTraceOut {avar name var idx op} {
    # Propagate changes in the matrix data array into the linked array.

    variable ${name}::unset

    if {![string compare $avar $unset]} {
	# Do not change the variable currently unsetting
	# one of its elements.
	return
    }

    variable ${name}::lock
    set lock 1 ; # Disable MatTraceIn [#532783]

    upvar #0 $avar    array
    variable ${name}::data
    variable ${name}::link

    set transpose $link($avar)

    if {$transpose} {
	foreach {r c} [split $idx ,] break
    } else {
	foreach {c r} [split $idx ,] break
    }

    if {$unset != {}} {
	# We are currently propagating the unset of an
	# element in a different linked array to this
	# array. We make sure that this is an unset too.

	unset array($c,$r)
    } else {
	set array($c,$r) $data($idx)
    }
    set lock 0
    return
}

# ::struct::matrix::SortMaxHeapify --
#
#	Helper for the 'sort' method. Performs the central algorithm
#	which converts the matrix into a heap, easily sortable.
#
# Arguments:
#	name	Matrix object which is sorted.
#	i	Index of the row/column currently being sorted.
#	key	Index of the column/row to sort the rows/columns by.
#	rowCol	Indicator if we are sorting rows ('r'), or columns ('c').
#	heapSize Number of rows/columns to sort.
#	rev	Boolean flag, set if sorting is done revers (-decreasing).
#
# Sideeffects:
#	Transforms the matrix into a heap of rows/columns,
#	swapping them around.
#
# Results:
#	None.

proc ::struct::matrix::SortMaxHeapify {name i key rowCol heapSize {rev 0}} {
    # MAX-HEAPIFY, adapted by EAS from CLRS 6.2
    switch  $rowCol {
	r { set A [GetColumn $name $key] }
	c { set A [GetRow    $name $key] }
    }
    # Weird expressions below for clarity, as CLRS uses A[1...n]
    # format and TCL uses A[0...n-1]
    set left  [expr {int(2*($i+1)    -1)}]
    set right [expr {int(2*($i+1)+1  -1)}]

    # left, right are tested as < rather than <= because they are
    # in A[0...n-1]
    if {
	$left < $heapSize &&
	( !$rev && [lindex $A $left] > [lindex $A $i] ||
	   $rev && [lindex $A $left] < [lindex $A $i] )
    } {
	set largest $left
    } else {
	set largest $i
    }

    if {
	$right < $heapSize &&
	( !$rev && [lindex $A $right] > [lindex $A $largest] ||
	   $rev && [lindex $A $right] < [lindex $A $largest] )
    } {
	set largest $right
    }

    if { $largest != $i } {
	switch $rowCol {
	    r { SwapRows    $name $i $largest }
	    c { SwapColumns $name $i $largest }
	}
	SortMaxHeapify $name $largest $key $rowCol $heapSize $rev
    }
    return
}

# ::struct::matrix::CheckSerialization --
#
#	Validate the serialization of a matrix.
#
# Arguments:
#	ser	Serialization to validate.
#	rvar	Variable to store the number of rows into.
#	cvar	Variable to store the number of columns into.
#	dvar	Variable to store the matrix data into.
#
# Results:
#	none

proc ::struct::matrix::CheckSerialization {ser rvar cvar dvar} {
    upvar 1 \
	    $rvar   rows    \
	    $cvar   columns \
	    $dvar   data

    # Overall length ok ?
    if {[llength $ser] != 3} {
	return -code error \
		"error in serialization: list length not 3."
    }

    foreach {r c d} $ser break

    # Check rows/columns information

    if {![string is integer -strict $r] || ($r < 0)} {
	return -code error \
		"error in serialization: bad number of rows \"$r\"."
    }
    if {![string is integer -strict $c] || ($c < 0)} {
	return -code error \
		"error in serialization: bad number of columns \"$c\"."
    }

    # Validate data against rows/columns. We can have less data than
    # rows or columns, the missing cells will be initialized to the
    # empty string. But too many is considered as a signal of
    # being something wrong.

    if {[llength $d] > $r} {
	return -code error \
		"error in serialization: data for to many rows."	
    }
    foreach rv $d {
	if {[llength $rv] > $c} {
	    return -code error \
		    "error in serialization: data for to many columns."	
	}
    }

    # Ok. The data is now ready for the caller.

    set data    $d
    set rows    $r
    set columns $c
    return
}

# ::struct::matrix::DeleteRows --
#
#	Deletes n rows from the bottom of the matrix.
#
# Arguments:
#	name	Name of the matrix.
#	n	The number of rows to delete (no greater than the number of rows).
#
# Results:
#	None.

proc ::struct::matrix::DeleteRows {name n} {
    variable ${name}::data
    variable ${name}::rows
    variable ${name}::columns
    variable ${name}::colw
    variable ${name}::rowh

    # Move all data from the higher rows down and then delete the
    # superfluous data in the old last row. Move the data in the
    # height cache too, take partial fill into account there too.
    # Invalidate the width cache for all columns.

    set rowstart [expr {$rows - $n}]

    for {set c 0} {$c < $columns} {incr c} {
	for {set r $rowstart} {$r < $rows} {incr r} {
	    unset data($c,$r)
	    catch {unset rowh($r)}
	}
	catch {unset colw($c)}
    }
    set rows $rowstart
    return
}

# ::struct::matrix::DeleteColumns --
#
#	Deletes n columns from the right of the matrix.
#
# Arguments:
#	name	Name of the matrix.
#	n	The number of columns to delete.
#
# Results:
#	None.

proc ::struct::matrix::DeleteColumns {name n} {
    variable ${name}::data
    variable ${name}::rows
    variable ${name}::columns
    variable ${name}::colw
    variable ${name}::rowh

    # Move all data from the higher columns down and then delete the
    # superfluous data in the old last column. Move the data in the
    # width cache too, take partial fill into account there too.
    # Invalidate the height cache for all rows.

    set colstart [expr {$columns - $n}]

    for {set r 0} {$r < $rows} {incr r} {
	for {set c $colstart} {$c < $columns} {incr c} {
	    unset data($c,$r)
	    catch {unset colw($c)}
	}
	catch {unset rowh($r)}
    }
    set columns $colstart
    return
}


# ### ### ### ######### ######### #########
## Ready

namespace eval ::struct {
    # Get 'matrix::matrix' into the general structure namespace.
    namespace import -force matrix::matrix
    namespace export matrix
}
package provide struct::matrix 2.0.2
#source ini.tcl
# ini.tcl --
#
#       Querying and modifying old-style windows configuration files (.ini)
#
# Copyright (c) 2003-2007    Aaron Faupell <afaupell@users.sourceforge.net>
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Import: ini.tcl,v 1.15 2008/05/11 00:53:58 andreas_kupries Exp $

package provide inifile 0.2.3

namespace eval ini {
    variable nexthandle  0
    variable commentchar \;
}

proc ::ini::open {ini {mode r+}} {
    variable nexthandle

    if { ![regexp {^(w|r)\+?$} $mode] } {
        error "$mode is not a valid access mode"
    }

    ::set fh ini$nexthandle
    ::set tmp [::open $ini $mode]
    fconfigure $tmp -translation crlf

    namespace eval ::ini::$fh {
        variable data;     array set data     {}
        variable comments; array set comments {}
        variable sections; array set sections {}
    }
    ::set ::ini::${fh}::channel $tmp
    ::set ::ini::${fh}::file    [_normalize $ini]
    ::set ::ini::${fh}::mode    $mode

    incr nexthandle
    if { [string match "r*" $mode] } {
        _loadfile $fh
    }
    return $fh
}

# close the file and delete all stored info about it
# this does not save any changes. see ::ini::commit

proc ::ini::close {fh} {
    _valid_ns $fh
    ::close [::set ::ini::${fh}::channel]
    namespace delete ::ini::$fh
}

# write all changes to disk

proc ::ini::commit {fh} {
    _valid_ns $fh
    namespace eval ::ini::$fh {
        if { $mode == "r" } {
            error "cannot write to read-only file"
        }
        ::close $channel
        ::set channel [::open $file w]
        ::set char $::ini::commentchar
        #seek $channel 0 start
        foreach sec [array names sections] {
            if { [info exists comments($sec)] } {
                puts $channel "$char [join $comments($sec) "\n$char "]\n"
            }
            puts $channel "\[$sec\]"
            foreach key [lsort -dictionary [array names data [::ini::_globescape $sec]\000*]] {
                ::set key [lindex [split $key \000] 1]
                if {[info exists comments($sec\000$key)]} {
                    puts $channel "$char [join $comments($sec\000$key) "\n$char "]"
                }
                puts $channel "$key=$data($sec\000$key)"
            }
            puts $channel ""
        }
        catch { unset char sec key }
        close $channel
        ::set channel [::open $file r+]
    }
    return
}

# internal command to read in a file
# see open and revert for public commands

proc ::ini::_loadfile {fh} {
    namespace eval ::ini::$fh {
        ::set cur {}
        ::set com {}
        set char $::ini::commentchar
        seek $channel 0 start

        foreach line [split [read $channel] "\n"] {
            if { [string match "$char*" $line] } {
                lappend com [string trim [string range $line [string length $char] end]]
            } elseif { [string match {\[*\]} $line] } {
                ::set cur [string range $line 1 end-1]
                if { $cur == "" } { continue }
                ::set sections($cur) 1
                if { $com != "" } {
                    ::set comments($cur) $com
                    ::set com {}
                }
            } elseif { [string match {*=*} $line] } {
                ::set line [split $line =]
                ::set key [string trim [lindex $line 0]]
                if { $key == "" || $cur == "" } { continue }
                ::set value [string trim [join [lrange $line 1 end] =]]
                if { [regexp "^(\".*\")\s+${char}(.*)$" $value -> 1 2] } {
                    set value $1
                    lappend com $2
                }
                ::set data($cur\000$key) $value
                if { $com != "" } {
                    ::set comments($cur\000$key) $com
                    ::set com {}
                }
            }
        }
        unset char cur com
        catch { unset line key value 1 2 }
    }
}

# internal command to escape glob special characters

proc ::ini::_globescape {string} {
    return [string map {* \\* ? \\? \\ \\\\ \[ \\\[ \] \\\]} $string]
}

# internal command to check if a section or key is nonexistant

proc ::ini::_exists {fh sec args} {
    if { ![info exists ::ini::${fh}::sections($sec)] } {
        error "no such section \"$sec\""
    }
    if { [llength $args] > 0 } {
        ::set key [lindex $args 0]
        if { ![info exists ::ini::${fh}::data($sec\000$key)] } {
            error "can't read key \"$key\""
        }
    }
}

# internal command to check validity of a handle

if { [package vcompare [package provide Tcl] 8.4] < 0 } {
    proc ::ini::_normalize {path} {
	return $path
    }
    proc ::ini::_valid_ns {name} {
	variable ::ini::${name}::data
	if { ![info exists data] } {
	    error "$name is not an open INI file"
	}
    }
} else {
    proc ::ini::_normalize {path} {
	file normalize $path
    }
    proc ::ini::_valid_ns {name} {
	if { ![namespace exists ::ini::$name] } {
	    error "$name is not an open INI file"
	}
    }
}

# get and set the ini comment character

proc ::ini::commentchar { {new {}} } {
    variable commentchar
    if {$new != ""} {
        if {[string length $new] > 1} {
	    return -code error "comment char must be a single character"
	}
        ::set commentchar $new
    }
    return $commentchar
}

# return all section names

proc ::ini::sections {fh} {
    _valid_ns $fh
    return [array names ::ini::${fh}::sections]
}

# return boolean indicating existance of section or key in section

proc ::ini::exists {fh sec {key {}}} {
    _valid_ns $fh
    if { $key == "" } {
        return [info exists ::ini::${fh}::sections($sec)]
    }
    return [info exists ::ini::${fh}::data($sec\000$key)]
}

# return all key names of section
# error if section is nonexistant

proc ::ini::keys {fh sec} {
    _valid_ns $fh
    _exists $fh $sec
    ::set keys {}
    foreach x [array names ::ini::${fh}::data [_globescape $sec]\000*] {
        lappend keys [lindex [split $x \000] 1]
    }
    return $keys
}

# return all key value pairs of section
# error if section is nonexistant

proc ::ini::get {fh sec} {
    _valid_ns $fh
    _exists $fh $sec
    upvar 0 ::ini::${fh}::data data
    ::set r {}
    foreach x [array names data [_globescape $sec]\000*] {
        lappend r [lindex [split $x \000] 1] $data($x)
    }
    return $r
}

# return the value of a key
# return default value if key or section is nonexistant otherwise error

proc ::ini::value {fh sec key {default {}}} {
    _valid_ns $fh
    if {$default != "" && ![info exists ::ini::${fh}::data($sec\000$key)]} {
        return $default
    }
    _exists $fh $sec $key
    return [::set ::ini::${fh}::data($sec\000$key)]
}

# set the value of a key
# new section or key names are created

proc ::ini::set {fh sec key value} {
    _valid_ns $fh
    ::set sec [string trim $sec]
    ::set key [string trim $key]
    if { $sec == "" || $key == "" } {
        error "section or key may not be empty"
    }
    ::set ::ini::${fh}::data($sec\000$key) $value
    ::set ::ini::${fh}::sections($sec) 1
    return $value
}

# delete a key or an entire section
# may delete nonexistant keys and sections

proc ::ini::delete {fh sec {key {}}} {
    _valid_ns $fh
    if { $key == "" } {
        array unset ::ini::${fh}::data [_globescape $sec]\000*
        array unset ::ini::${fh}::sections [_globescape $sec]
    }
    catch {unset ::ini::${fh}::data($sec\000$key)}
}

# read and set comments for sections and keys
# may comment nonexistant sections and keys

proc ::ini::comment {fh sec key args} {
    _valid_ns $fh
    upvar 0 ::ini::${fh}::comments comments
    ::set r $sec
    if { $key != "" } { append r \000$key }
    if { [llength $args] == 0 } {
        if { ![info exists comments($r)] } { return {} }
        return $comments($r)
    }
    if { [llength $args] == 1 && [lindex $args 0] == "" } {
        unset -nocomplain comments($r)
        return {}
    }
    # take care of any embedded newlines
    for {::set i 0} {$i < [llength $args]} {incr i} {
        ::set args [eval [list lreplace $args $i $i] [split [lindex $args $i] \n]]
    }
    eval [list lappend comments($r)] $args
}

# return the physical filename for the handle

proc ::ini::filename {fh} {
    _valid_ns $fh
    return [::set ::ini::${fh}::file]
}

# reload the file from disk losing all changes since the last commit

proc ::ini::revert {fh} {
    _valid_ns $fh
    namespace eval ::ini::$fh {
        array set data     {}
        array set comments {}
        array set sections {}
    }
    if { ![string match "w*" $mode] } {
        _loadfile $fh
    }
}
###############################################################################
# cmdline.tcl --
#
#	This package provides a utility for parsing command line
#	arguments that are processed by our various applications.
#	It also includes a utility routine to determine the
#	application name for use in command line errors.
#
# Copyright (c) 1998-2000 by Ajuba Solutions.
# Copyright (c) 2001-2006 by Andreas Kupries <andreas_kupries@users.sf.net>.
# Copyright (c) 2003      by David N. Welton  <davidw@dedasys.com>
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Import: cmdline.tcl,v 1.26 2008/07/09 18:02:59 andreas_kupries Exp $

package provide cmdline 1.3.1

namespace eval ::cmdline {
    namespace export getArgv0 getopt getKnownOpt getfiles getoptions \
	    getKnownOptions usage
}

# ::cmdline::getopt --
#
#	The cmdline::getopt works in a fashion like the standard
#	C based getopt function.  Given an option string and a 
#	pointer to an array or args this command will process the
#	first argument and return info on how to proceed.
#
# Arguments:
#	argvVar		Name of the argv list that you
#			want to process.  If options are found the
#			arg list is modified and the processed arguments
#			are removed from the start of the list.
#	optstring	A list of command options that the application
#			will accept.  If the option ends in ".arg" the
#			getopt routine will use the next argument as 
#			an argument to the option.  Otherwise the option	
#			is a boolean that is set to 1 if present.
#	optVar		The variable pointed to by optVar
#			contains the option that was found (without the
#			leading '-' and without the .arg extension).
#	valVar		Upon success, the variable pointed to by valVar
#			contains the value for the specified option.
#			This value comes from the command line for .arg
#			options, otherwise the value is 1.
#			If getopt fails, the valVar is filled with an
#			error message.
#
# Results:
# 	The getopt function returns 1 if an option was found, 0 if no more
# 	options were found, and -1 if an error occurred.

proc ::cmdline::getopt {argvVar optstring optVar valVar} {
    upvar 1 $argvVar argsList
    upvar 1 $optVar option
    upvar 1 $valVar value

    set result [getKnownOpt argsList $optstring option value]

    if {$result < 0} {
        # Collapse unknown-option error into any-other-error result.
        set result -1
    }
    return $result
}

# ::cmdline::getKnownOpt --
#
#	The cmdline::getKnownOpt works in a fashion like the standard
#	C based getopt function.  Given an option string and a 
#	pointer to an array or args this command will process the
#	first argument and return info on how to proceed.
#
# Arguments:
#	argvVar		Name of the argv list that you
#			want to process.  If options are found the
#			arg list is modified and the processed arguments
#			are removed from the start of the list.  Note that
#			unknown options and the args that follow them are
#			left in this list.
#	optstring	A list of command options that the application
#			will accept.  If the option ends in ".arg" the
#			getopt routine will use the next argument as 
#			an argument to the option.  Otherwise the option	
#			is a boolean that is set to 1 if present.
#	optVar		The variable pointed to by optVar
#			contains the option that was found (without the
#			leading '-' and without the .arg extension).
#	valVar		Upon success, the variable pointed to by valVar
#			contains the value for the specified option.
#			This value comes from the command line for .arg
#			options, otherwise the value is 1.
#			If getopt fails, the valVar is filled with an
#			error message.
#
# Results:
# 	The getKnownOpt function returns 1 if an option was found,
#	0 if no more options were found, -1 if an unknown option was
#	encountered, and -2 if any other error occurred. 

proc ::cmdline::getKnownOpt {argvVar optstring optVar valVar} {
    upvar 1 $argvVar argsList
    upvar 1 $optVar  option
    upvar 1 $valVar  value

    # default settings for a normal return
    set value ""
    set option ""
    set result 0

    # check if we're past the end of the args list
    if {[llength $argsList] != 0} {

	# if we got -- or an option that doesn't begin with -, return (skipping
	# the --).  otherwise process the option arg.
	switch -glob -- [set arg [lindex $argsList 0]] {
	    "--" {
		set argsList [lrange $argsList 1 end]
	    }

	    "-*" {
		set option [string range $arg 1 end]

		if {[lsearch -exact $optstring $option] != -1} {
		    # Booleans are set to 1 when present
		    set value 1
		    set result 1
		    set argsList [lrange $argsList 1 end]
		} elseif {[lsearch -exact $optstring "$option.arg"] != -1} {
		    set result 1
		    set argsList [lrange $argsList 1 end]
		    if {[llength $argsList] != 0} {
			set value [lindex $argsList 0]
			set argsList [lrange $argsList 1 end]
		    } else {
			set value "Option \"$option\" requires an argument"
			set result -2
		    }
		} else {
		    # Unknown option.
		    set value "Illegal option \"-$option\""
		    set result -1
		}
	    }
	    default {
		# Skip ahead
	    }
	}
    }

    return $result
}

# ::cmdline::getoptions --
#
#	Process a set of command line options, filling in defaults
#	for those not specified.  This also generates an error message
#	that lists the allowed flags if an incorrect flag is specified.
#
# Arguments:
#	arglistVar	The name of the argument list, typically argv.
#			We remove all known options and their args from it.
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#				(where flag takes no argument) 
#					flag comment 
#
#				(or where flag takes an argument) 
#					flag default comment
#
#			If flag ends in ".arg" then the value is taken from the
#			command line. Otherwise it is a boolean and appears in
#			the result if present on the command line. If flag ends
#			in ".secret", it will not be displayed in the usage.
#	usage		Text to include in the usage display. Defaults to
#			"options:"
#
# Results
#	Name value pairs suitable for using with array set.

proc ::cmdline::getoptions {arglistVar optlist {usage options:}} {
    upvar 1 $arglistVar argv

    set opts [GetOptionDefaults $optlist result]

    set argc [llength $argv]
    while {[set err [getopt argv $opts opt arg]]} {
	if {$err < 0} {
            set result(?) ""
            break
	}
	set result($opt) $arg
    }
    if {[info exist result(?)] || [info exists result(help)]} {
	error [usage $optlist $usage]
    }
    return [array get result]
}

# ::cmdline::getKnownOptions --
#
#	Process a set of command line options, filling in defaults
#	for those not specified.  This ignores unknown flags, but generates
#	an error message that lists the correct usage if a known option
#	is used incorrectly.
#
# Arguments:
#	arglistVar	The name of the argument list, typically argv.  This
#			We remove all known options and their args from it.
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#				flag default comment
#			If flag ends in ".arg" then the value is taken from the
#			command line. Otherwise it is a boolean and appears in
#			the result if present on the command line. If flag ends
#			in ".secret", it will not be displayed in the usage.
#	usage		Text to include in the usage display. Defaults to
#			"options:"
#
# Results
#	Name value pairs suitable for using with array set.

proc ::cmdline::getKnownOptions {arglistVar optlist {usage options:}} {
    upvar 1 $arglistVar argv

    set opts [GetOptionDefaults $optlist result]

    # As we encounter them, keep the unknown options and their
    # arguments in this list.  Before we return from this procedure,
    # we'll prepend these args to the argList so that the application
    # doesn't lose them.

    set unknownOptions [list]

    set argc [llength $argv]
    while {[set err [getKnownOpt argv $opts opt arg]]} {
	if {$err == -1} {
            # Unknown option.

            # Skip over any non-option items that follow it.
            # For now, add them to the list of unknownOptions.
            lappend unknownOptions [lindex $argv 0]
            set argv [lrange $argv 1 end]
            while {([llength $argv] != 0) \
                    && ![string match "-*" [lindex $argv 0]]} {
                lappend unknownOptions [lindex $argv 0]
                set argv [lrange $argv 1 end]
            }
	} elseif {$err == -2} {
            set result(?) ""
            break
        } else {
            set result($opt) $arg
        }
    }

    # Before returning, prepend the any unknown args back onto the
    # argList so that the application doesn't lose them.
    set argv [concat $unknownOptions $argv]

    if {[info exist result(?)] || [info exists result(help)]} {
	error [usage $optlist $usage]
    }
    return [array get result]
}

# ::cmdline::GetOptionDefaults --
#
#	This internal procedure processes the option list (that was passed to
#	the getopt or getKnownOpt procedure).  The defaultArray gets an index
#	for each option in the option list, the value of which is the option's
#	default value.
#
# Arguments:
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#				flag default comment
#			If flag ends in ".arg" then the value is taken from the
#			command line. Otherwise it is a boolean and appears in
#			the result if present on the command line. If flag ends
#			in ".secret", it will not be displayed in the usage.
#	defaultArrayVar	The name of the array in which to put argument defaults.
#
# Results
#	Name value pairs suitable for using with array set.

proc ::cmdline::GetOptionDefaults {optlist defaultArrayVar} {
    upvar 1 $defaultArrayVar result

    set opts {? help}
    foreach opt $optlist {
	set name [lindex $opt 0]
	if {[regsub -- .secret$ $name {} name] == 1} {
	    # Need to hide this from the usage display and getopt
	}   
	lappend opts $name
	if {[regsub -- .arg$ $name {} name] == 1} {

	    # Set defaults for those that take values.

	    set default [lindex $opt 1]
	    set result($name) $default
	} else {
	    # The default for booleans is false
	    set result($name) 0
	}
    }
    return $opts
}

# ::cmdline::usage --
#
#	Generate an error message that lists the allowed flags.
#
# Arguments:
#	optlist		As for cmdline::getoptions
#	usage		Text to include in the usage display. Defaults to
#			"options:"
#
# Results
#	A formatted usage message

proc ::cmdline::usage {optlist {usage {options:}}} {
    set str "[getArgv0] $usage\n"
    foreach opt [concat $optlist \
	    {{help "Print this message"} {? "Print this message"}}] {
	set name [lindex $opt 0]
	if {[regsub -- .secret$ $name {} name] == 1} {
	    # Hidden option
	    continue
	}
	if {[regsub -- .arg$ $name {} name] == 1} {
	    set default [lindex $opt 1]
	    set comment [lindex $opt 2]
	    append str [format " %-20s %s <%s>\n" "-$name value" \
		    $comment $default]
	} else {
	    set comment [lindex $opt 1]
	    append str [format " %-20s %s\n" "-$name" $comment]
	}
    }
    return $str
}

# ::cmdline::getfiles --
#
#	Given a list of file arguments from the command line, compute
#	the set of valid files.  On windows, file globbing is performed
#	on each argument.  On Unix, only file existence is tested.  If
#	a file argument produces no valid files, a warning is optionally
#	generated.
#
#	This code also uses the full path for each file.  If not
#	given it prepends [pwd] to the filename.  This ensures that
#	these files will never conflict with files in our zip file.
#
# Arguments:
#	patterns	The file patterns specified by the user.
#	quiet		If this flag is set, no warnings will be generated.
#
# Results:
#	Returns the list of files that match the input patterns.

proc ::cmdline::getfiles {patterns quiet} {
    set result {}
    if {$::tcl_platform(platform) == "windows"} {
	foreach pattern $patterns {
	    set pat [file join $pattern]
	    set files [glob -nocomplain -- $pat]
	    if {$files == {}} {
		if {! $quiet} {
		    puts stdout "warning: no files match \"$pattern\""
		}
	    } else {
		foreach file $files {
		    lappend result $file
		}
	    }
	}
    } else {
	set result $patterns
    }
    set files {}
    foreach file $result {
	# Make file an absolute path so that we will never conflict
	# with files that might be contained in our zip file.
	set fullPath [file join [pwd] $file]
	
	if {[file isfile $fullPath]} {
	    lappend files $fullPath
	} elseif {! $quiet} {
	    puts stdout "warning: no files match \"$file\""
	}
    }
    return $files
}

# ::cmdline::getArgv0 --
#
#	This command returns the "sanitized" version of argv0.  It will strip
#	off the leading path and remove the ".bin" extensions that our apps
#	use because they must be wrapped by a shell script.
#
# Arguments:
#	None.
#
# Results:
#	The application name that can be used in error messages.

proc ::cmdline::getArgv0 {} {
    global argv0

    set name [file tail $argv0]
    return [file rootname $name]
}

##
# ### ### ### ######### ######### #########
##
# Now the typed versions of the above commands.
##
# ### ### ### ######### ######### #########
##

# typedCmdline.tcl --
#
#    This package provides a utility for parsing typed command
#    line arguments that may be processed by various applications.
#
# Copyright (c) 2000 by Ross Palmer Mohn.
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Import: cmdline.tcl,v 1.26 2008/07/09 18:02:59 andreas_kupries Exp $

namespace eval ::cmdline {
    namespace export typedGetopt typedGetoptions typedUsage

    # variable cmdline::charclasses --
    #
    #    Create regexp list of allowable character classes
    #    from "string is" error message.
    #
    # Results:
    #    String of character class names separated by "|" characters.

    variable charclasses
    #checker exclude badKey
    catch {string is . .} charclasses
    variable dummy
    regexp      -- {must be (.+)$} $charclasses dummy charclasses
    regsub -all -- {, (or )?}      $charclasses {|}   charclasses
    unset dummy
}

# ::cmdline::typedGetopt --
#
#	The cmdline::typedGetopt works in a fashion like the standard
#	C based getopt function.  Given an option string and a
#	pointer to a list of args this command will process the
#	first argument and return info on how to proceed. In addition,
#	you may specify a type for the argument to each option.
#
# Arguments:
#	argvVar		Name of the argv list that you want to process.
#			If options are found, the arg list is modified
#			and the processed arguments are removed from the
#			start of the list.
#
#	optstring	A list of command options that the application
#			will accept.  If the option ends in ".xxx", where
#			xxx is any valid character class to the tcl
#			command "string is", then typedGetopt routine will
#			use the next argument as a typed argument to the
#			option. The argument must match the specified
#			character classes (e.g. integer, double, boolean,
#			xdigit, etc.). Alternatively, you may specify
#			".arg" for an untyped argument.
#
#	optVar		Upon success, the variable pointed to by optVar
#			contains the option that was found (without the
#			leading '-' and without the .xxx extension).  If
#			typedGetopt fails the variable is set to the empty
#			string. SOMETIMES! Different for each -value!
#
#	argVar		Upon success, the variable pointed to by argVar
#			contains the argument for the specified option.
#			If typedGetopt fails, the variable is filled with
#			an error message.
#
# Argument type syntax:
#	Option that takes no argument.
#		foo
#
#	Option that takes a typeless argument.
#		foo.arg
#
#	Option that takes a typed argument. Allowable types are all
#	valid character classes to the tcl command "string is".
#	Currently must be one of alnum, alpha, ascii, control,
#	boolean, digit, double, false, graph, integer, lower, print,
#	punct, space, true, upper, wordchar, or xdigit.
#		foo.double
#
#	Option that takes an argument from a list.
#		foo.(bar|blat)
#
# Argument quantifier syntax:
#	Option that takes an optional argument.
#		foo.arg?
#
#	Option that takes a list of arguments terminated by "--".
#		foo.arg+
#
#	Option that takes an optional list of arguments terminated by "--".
#		foo.arg*
#
#	Argument quantifiers work on all argument types, so, for
#	example, the following is a valid option specification.
#		foo.(bar|blat|blah)?
#
# Argument syntax miscellany:
#	Options may be specified on the command line using a unique,
#	shortened version of the option name. Given that program foo
#	has an option list of {bar.alpha blah.arg blat.double},
#	"foo -b fob" returns an error, but "foo -ba fob"
#	successfully returns {bar fob}
#
# Results:
#	The typedGetopt function returns one of the following:
#	 1	a valid option was found
#	 0	no more options found to process
#	-1	invalid option
#	-2	missing argument to a valid option
#	-3	argument to a valid option does not match type
#
# Known Bugs:
#	When using options which include special glob characters,
#	you must use the exact option. Abbreviating it can cause
#	an error in the "cmdline::prefixSearch" procedure.

proc ::cmdline::typedGetopt {argvVar optstring optVar argVar} {
    variable charclasses

    upvar $argvVar argsList

    upvar $optVar retvar
    upvar $argVar optarg

    # default settings for a normal return
    set optarg ""
    set retvar ""
    set retval 0

    # check if we're past the end of the args list
    if {[llength $argsList] != 0} {

        # if we got -- or an option that doesn't begin with -, return (skipping
        # the --).  otherwise process the option arg.
        switch -glob -- [set arg [lindex $argsList 0]] {
            "--" {
                set argsList [lrange $argsList 1 end]
            }

            "-*" {
                # Create list of options without their argument extensions

                set optstr ""
                foreach str $optstring {
                    lappend optstr [file rootname $str]
                }

                set _opt [string range $arg 1 end]

                set i [prefixSearch $optstr [file rootname $_opt]]
                if {$i != -1} {
                    set opt [lindex $optstring $i]

                    set quantifier "none"
                    if {[regexp -- {\.[^.]+([?+*])$} $opt dummy quantifier]} {
                        set opt [string range $opt 0 end-1]
                    }

                    if {[string first . $opt] == -1} {
                        set retval 1
                        set retvar $opt
                        set argsList [lrange $argsList 1 end]

                    } elseif {[regexp -- "\\.(arg|$charclasses)\$" $opt dummy charclass]
                            || [regexp -- {\.\(([^)]+)\)} $opt dummy charclass]} {
				if {[string equal arg $charclass]} {
                            set type arg
			} elseif {[regexp -- "^($charclasses)\$" $charclass]} {
                            set type class
                        } else {
                            set type oneof
                        }

                        set argsList [lrange $argsList 1 end]
                        set opt [file rootname $opt]

                        while {1} {
                            if {[llength $argsList] == 0
                                    || [string equal "--" [lindex $argsList 0]]} {
                                if {[string equal "--" [lindex $argsList 0]]} {
                                    set argsList [lrange $argsList 1 end]
                                }

                                set oneof ""
                                if {$type == "arg"} {
                                    set charclass an
                                } elseif {$type == "oneof"} {
                                    set oneof ", one of $charclass"
                                    set charclass an
                                }
    
                                if {$quantifier == "?"} {
                                    set retval 1
                                    set retvar $opt
                                    set optarg ""
                                } elseif {$quantifier == "+"} {
                                    set retvar $opt
                                    if {[llength $optarg] < 1} {
                                        set retval -2
                                        set optarg "Option requires at least one $charclass argument$oneof -- $opt"
                                    } else {
                                        set retval 1
                                    }
                                } elseif {$quantifier == "*"} {
                                    set retval 1
                                    set retvar $opt
                                } else {
                                    set optarg "Option requires $charclass argument$oneof -- $opt"
                                    set retvar $opt
                                    set retval -2
                                }
                                set quantifier ""
                            } elseif {($type == "arg")
                                    || (($type == "oneof")
                                    && [string first "|[lindex $argsList 0]|" "|$charclass|"] != -1)
                                    || (($type == "class")
                                    && [string is $charclass [lindex $argsList 0]])} {
                                set retval 1
                                set retvar $opt
                                lappend optarg [lindex $argsList 0]
                                set argsList [lrange $argsList 1 end]
                            } else {
                                set oneof ""
                                if {$type == "arg"} {
                                    set charclass an
                                } elseif {$type == "oneof"} {
                                    set oneof ", one of $charclass"
                                    set charclass an
                                }
                                set optarg "Option requires $charclass argument$oneof -- $opt"
                                set retvar $opt
                                set retval -3
    
                                if {$quantifier == "?"} {
                                    set retval 1
                                    set optarg ""
                                }
                                set quantifier ""
                            }
                             if {![regexp -- {[+*]} $quantifier]} {
                                break;
                            }
                        }
                    } else {
                        error "Illegal option type specification:\
                                must be one of $charclasses"
                    }
                } else {
                    set optarg "Illegal option -- $_opt"
                    set retvar $_opt
                    set retval -1
                }
            }
	    default {
		# Skip ahead
	    }
        }
    }

    return $retval
}

# ::cmdline::typedGetoptions --
#
#	Process a set of command line options, filling in defaults
#	for those not specified. This also generates an error message
#	that lists the allowed options if an incorrect option is
#	specified.
#
# Arguments:
#	arglistVar	The name of the argument list, typically argv
#	optlist		A list-of-lists where each element specifies an option
#			in the form:
#
#				option default comment
#
#			Options formatting is as described for the optstring
#			argument of typedGetopt. Default is for optionally
#			specifying a default value. Comment is for optionally
#			specifying a comment for the usage display. The
#			options "-help" and "-?" are automatically included
#			in optlist.
#
# Argument syntax miscellany:
#	Options formatting and syntax is as described in typedGetopt.
#	There are two additional suffixes that may be applied when
#	passing options to typedGetoptions.
#
#	You may add ".multi" as a suffix to any option. For options
#	that take an argument, this means that the option may be used
#	more than once on the command line and that each additional
#	argument will be appended to a list, which is then returned
#	to the application.
#		foo.double.multi
#
#	If a non-argument option is specified as ".multi", it is
#	toggled on and off for each time it is used on the command
#	line.
#		foo.multi
#
#	If an option specification does not contain the ".multi"
#	suffix, it is not an error to use an option more than once.
#	In this case, the behavior for options with arguments is that
#	the last argument is the one that will be returned. For
#	options that do not take arguments, using them more than once
#	has no additional effect.
#
#	Options may also be hidden from the usage display by
#	appending the suffix ".secret" to any option specification.
#	Please note that the ".secret" suffix must be the last suffix,
#	after any argument type specification and ".multi" suffix.
#		foo.xdigit.multi.secret
#
# Results
#	Name value pairs suitable for using with array set.

proc ::cmdline::typedGetoptions {arglistVar optlist {usage options:}} {
    variable charclasses

    upvar 1 $arglistVar argv

    set opts {? help}
    foreach opt $optlist {
        set name [lindex $opt 0]
        if {[regsub -- {\.secret$} $name {} name] == 1} {
            # Remove this extension before passing to typedGetopt.
        }
        if {[regsub -- {\.multi$} $name {} name] == 1} {
            # Remove this extension before passing to typedGetopt.

            regsub -- {\..*$} $name {} temp
            set multi($temp) 1
        }
        lappend opts $name
        if {[regsub -- "\\.(arg|$charclasses|\\(.+).?\$" $name {} name] == 1} {
            # Set defaults for those that take values.
            # Booleans are set just by being present, or not

            set dflt [lindex $opt 1]
            if {$dflt != {}} {
                set defaults($name) $dflt
            }
        }
    }
    set argc [llength $argv]
    while {[set err [typedGetopt argv $opts opt arg]]} {
        if {$err == 1} {
            if {[info exists result($opt)]
                    && [info exists multi($opt)]} {
                # Toggle boolean options or append new arguments

                if {$arg == ""} {
                    unset result($opt)
                } else {
                    set result($opt) "$result($opt) $arg"
                }
            } else {
                set result($opt) "$arg"
            }
        } elseif {($err == -1) || ($err == -3)} {
            error [typedUsage $optlist $usage]
        } elseif {$err == -2 && ![info exists defaults($opt)]} {
            error [typedUsage $optlist $usage]
        }
    }
    if {[info exists result(?)] || [info exists result(help)]} {
        error [typedUsage $optlist $usage]
    }
    foreach {opt dflt} [array get defaults] {
        if {![info exists result($opt)]} {
            set result($opt) $dflt
        }
    }
    return [array get result]
}

# ::cmdline::typedUsage --
#
#	Generate an error message that lists the allowed flags,
#	type of argument taken (if any), default value (if any),
#	and an optional description.
#
# Arguments:
#	optlist		As for cmdline::typedGetoptions
#
# Results
#	A formatted usage message

proc ::cmdline::typedUsage {optlist {usage {options:}}} {
    variable charclasses

    set str "[getArgv0] $usage\n"
    foreach opt [concat $optlist \
            {{help "Print this message"} {? "Print this message"}}] {
        set name [lindex $opt 0]
        if {[regsub -- {\.secret$} $name {} name] == 1} {
            # Hidden option

        } else {
            if {[regsub -- {\.multi$} $name {} name] == 1} {
                # Display something about multiple options
            }

            if {[regexp -- "\\.(arg|$charclasses)\$" $name dummy charclass]
                    || [regexp -- {\.\(([^)]+)\)} $opt dummy charclass]} {
                   regsub -- "\\..+\$" $name {} name
                set comment [lindex $opt 2]
                set default "<[lindex $opt 1]>"
                if {$default == "<>"} {
                    set default ""
                }
                append str [format " %-20s %s %s\n" "-$name $charclass" \
                        $comment $default]
            } else {
                set comment [lindex $opt 1]
		append str [format " %-20s %s\n" "-$name" $comment]
            }
        }
    }
    return $str
}

# ::cmdline::prefixSearch --
#
#	Search a Tcl list for a pattern; searches first for an exact match,
#	and if that fails, for a unique prefix that matches the pattern 
#	(i.e, first "lsearch -exact", then "lsearch -glob $pattern*"
#
# Arguments:
#	list		list of words
#	pattern		word to search for
#
# Results:
#	Index of found word is returned. If no exact match or
#	unique short version is found then -1 is returned.

proc ::cmdline::prefixSearch {list pattern} {
    # Check for an exact match

    if {[set pos [::lsearch -exact $list $pattern]] > -1} {
        return $pos
    }

    # Check for a unique short version

    set slist [lsort $list]
    if {[set pos [::lsearch -glob $slist $pattern*]] > -1} {
        # What if there is nothing for the check variable?

        set check [lindex $slist [expr {$pos + 1}]]
        if {[string first $pattern $check] != 0} {
            return [::lsearch -exact $list [lindex $slist $pos]]
        }
    }
    return -1
}

# split.tcl --
#
#	Various ways of splitting a string.
#
# Copyright (c) 2000      by Ajuba Solutions.
# Copyright (c) 2000      by Eric Melski <ericm@ajubasolutions.com>
# Copyright (c) 2001      by Reinhard Max <max@suse.de>
# Copyright (c) 2003      by Pat Thoyts <patthoyts@users.sourceforge.net>
# Copyright (c) 2001-2006 by Andreas Kupries <andreas_kupries@users.sourceforge.net>
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.
# 
# RCS: @(#) $Import: split.tcl,v 1.7 2006/04/21 04:42:28 andreas_kupries Exp $

# ### ### ### ######### ######### #########
## Requirements


namespace eval ::textutil::split {}

########################################################################
# This one was written by Bob Techentin (RWT in Tcl'ers Wiki):
# http://www.techentin.net
# mailto:techentin.robert@mayo.edu
#
# Later, he send me an email stated that I can use it anywhere, because
# no copyright was added, so the code is defacto in the public domain.
#
# You can found it in the Tcl'ers Wiki here:
# http://mini.net/cgi-bin/wikit/460.html
#
# Bob wrote:
# If you need to split string into list using some more complicated rule
# than builtin split command allows, use following function. It mimics
# Perl split operator which allows regexp as element separator, but,
# like builtin split, it expects string to split as first arg and regexp
# as second (optional) By default, it splits by any amount of whitespace. 
# Note that if you add parenthesis into regexp, parenthesed part of separator
# would be added into list as additional element. Just like in Perl. -- cary 
#
# Speed improvement by Reinhard Max:
# Instead of repeatedly copying around the not yet matched part of the
# string, I use [regexp]'s -start option to restrict the match to that
# part. This reduces the complexity from something like O(n^1.5) to
# O(n). My test case for that was:
# 
# foreach i {1 10 100 1000 10000} {
#     set s [string repeat x $i]
#     puts [time {splitx $s .}]
# }
#

if {[package vsatisfies [package provide Tcl] 8.3]} {

    proc ::textutil::split::splitx {str {regexp {[\t \r\n]+}}} {
        # Bugfix 476988
        if {[string length $str] == 0} {
            return {}
        }
        if {[string length $regexp] == 0} {
            return [::split $str ""]
        }
        set list  {}
        set start 0
        while {[regexp -start $start -indices -- $regexp $str match submatch]} {
            foreach {subStart subEnd} $submatch break
            foreach {matchStart matchEnd} $match break
            incr matchStart -1
            incr matchEnd
            lappend list [string range $str $start $matchStart]
            if {$subStart >= $start} {
                lappend list [string range $str $subStart $subEnd]
            }
            set start $matchEnd
        }
        lappend list [string range $str $start end]
        return $list
    }

} else {    
    # For tcl <= 8.2 we do not have regexp -start...
    proc ::textutil::split::splitx [list str [list regexp "\[\t \r\n\]+"]] {

        if {[string length $str] == 0} {
            return {}
        }
        if {[string length $regexp] == 0} {
            return [::split $str {}]
        }

        set list  {}
        while {[regexp -indices -- $regexp $str match submatch]} {
            lappend list [string range $str 0 [expr {[lindex $match 0] -1}]]
            if {[lindex $submatch 0] >= 0} {
                lappend list [string range $str [lindex $submatch 0] \
                                  [lindex $submatch 1]]
            }
            set str [string range $str [expr {[lindex $match 1]+1}] end]
        }
        lappend list $str
        return $list
    }
    
}

#
# splitn --
#
# splitn splits the string $str into chunks of length $len.  These
# chunks are returned as a list.
#
# If $str really contains a ByteArray object (as retrieved from binary
# encoded channels) splitn must honor this by splitting the string
# into chunks of $len bytes.
#
# It is an error to call splitn with a nonpositive $len.
#
# If splitn is called with an empty string, it returns the empty list.
#
# If the length of $str is not an entire multiple of the chunk length,
# the last chunk in the generated list will be shorter than $len.
#
# The implementation presented here was given by Bryan Oakley, as
# part of a ``contest'' I staged on c.l.t in July 2004.  I selected
# this version, as it does not rely on runtime generated code, is
# very fast for chunk size one, not too bad in all the other cases,
# and uses [split] or [string range] which have been around for quite
# some time.
#
# -- Robert Suetterlin (robert@mpe.mpg.de)
#
proc ::textutil::split::splitn {str {len 1}} {

    if {$len <= 0} {
        return -code error "len must be > 0"
    }

    if {$len == 1} {
        return [split $str {}]
    }

    set result [list]
    set max [string length $str]
    set i 0
    set j [expr {$len -1}]
    while {$i < $max} {
        lappend result [string range $str $i $j]
        incr i $len
        incr j $len
    }

    return $result
}

# ### ### ### ######### ######### #########
## Data structures

namespace eval ::textutil::split {
    namespace export splitx splitn
}

# ### ### ### ######### ######### #########
## Ready

package provide textutil::split 0.7

###############################################################################
# JMXSTAT specific code starts here
###############################################################################
#array set params [::tcl::tm::roots]
#parray $params

# this is a set of stubs that are part of beastat and some will be built
# up to provide the equivalent functionality

proc get_server_to_test {args} {
  return 0
}
proc uptime_check {args} {
  return 0
}

proc init_status {arg} {
  switch -- $arg {
    Uptime { set result "60;30;>" }
    datacollection { set result "java.lang:type=ClassLoading java.lang:type=Threading java.lang:type=Memory java.lang:type=Runtime" }
    default { set result "" }
  }
  return $result
}

proc mydie { l } {
  global Options Usage
  puts stderr $l
  puts [::cmdline::usage $Options $Usage]
  exit 1
}

proc mylassign {sequence v args} {
  set args [linsert $args 0 $v]
  set a [::llength $args]

  # Perform assignments
  set i 0
  foreach v $args {
    upvar 1 $v var
    set var [::lindex $sequence $i]
    incr i
  }

  # Return remainder, if there is any.
  return [::lrange $sequence $a end]
}

proc debug_print {args} {
  global debug
  set msg "DBG: "
  foreach arg $args {
    set msg "$msg $arg"
  }
  if { $debug } {
    puts $msg
  }
}

proc query_bb { host co } {
#############################################################################
# In here we will fetch the current status log for the test <co> on the     #
# server or JVM <host>.  The utility to be used is a local binary, these    #
# binaries have different syntax depending on platform and implementation.  #
#############################################################################
  global XYMON XYMSRV XYMONCOMMS XYMONHOME OS S

  if { $OS == "Windows" } {

    # querying a status with BBWin
    # bbwincmd.exe <bbdisplay>[:<port>] query <hostname> <test>
    #
    # The output doesnt look a bit like a hobbitdlog output so do
    # our best to reformat it, the raw test will look like so:-
    #
    #bbwincmd 20.157.9.18 query cscsvradc007.spiderlink.local conn
    #bbdisplay defined to : 20.157.9.18
    #port defined to : 1984
    #Sending query ...
    #hostname defined to: cscsvradc007.spiderlink.local
    #
    #Sending to 20.157.9.18 :
    #query cscsvradc007.spiderlink.local.conn
    #
    #
    #green <!-- [flags:OrdAstLe] --> Mon Mar 21 17:18:42 2011 conn ok
    #
    #bbwincmd 20.157.9.18 query tomcat_fosportal_008 Threads
    #bbdisplay defined to : 20.157.9.18
    #port defined to : 1984
    #Sending query ...
    #hostname defined to: tomcat_fosportal_008
    #
    #Sending to 20.157.9.18 :
    #query tomcat_fosportal_008.Threads
    #
    #
    #green
    #
    #
    #
    set lines [ split [exec [ file join $XYMONHOME bin BBWinCmd.exe] $XYMSRV query $host $co] "\n"]
    set i 0
    set indata 0
    append result $host "|" $co "|"

    foreach line $lines {
      debug_print "Line $i/[llength $lines]: $line"
      regsub -all -- {\<!-- .* --\>} $line {} line
      if { [string match -nocase "query $host.$co" $line] } { set indata 1 }
      if { ($indata) && ( [regexp -nocase {(red|yellow|green|blue|clear|purple)} $line] ) } {
        regexp {^(red|yellow|green|blue|clear|purple).*$} $line -> l
        append result $l "|\n$l\n"
      }
      incr i
    }
    return $result

  } else {

    # querying a status with Xymon version 4.3
    # Usage: xymon [--debug] [--proxy=http://ip:port/] RECIPIENT DATA
    # RECIPIENT: IP-address, hostname or URL
    # DATA: Message to send, or "-" to read from stdin
    # The first line of the response contain a series of fields separated
    # by a pipe-sign:
    #   hostname The name of the host
    #   testname The name of the test
    #   color Status color (green, yellow, red, blue, clear, purple)
    #   .......
    # After the first line comes the full status log in plain text format. 
    set MESSAGE "hobbitdlog $host.$co"
    return [exec [ file join $XYMONHOME bin $XYMONCOMMS ] $XYMSRV $MESSAGE]

  }
}

proc send_bb { host co status MSG } {
#############################################################################
# In here we will take the page contained in MSG and send it to Xymon as a  #
# status update in a single message.  The host, column and status to which  #
# the update is to apply are given as arguments.  The utility to be used is #
# a local binary, these have different syntax depending on the platform and #
# implementation.                                                           #
#############################################################################
  global XYMON XYMSRV XYMONCOMMS OS S
  upvar #0 beanstart mystart
  upvar #0 params myparams
  #puts "sending $status message to Xymon about column $co on server $host"
  #need to clean dots from host and check XYMSRV, 0.0.0.0 will defer lookup
  #to the XYMON cmd, otherwise need to loop through IP addresses.

  if { $OS == "Windows" } {
    # Sending a status with BBWin
    # bbwincmd.exe <bbdisplay>[:<port>] status <hostname> <testname> <color> <text> [< lifetime>]
    if { $myparams(n) } {
      puts [exec $XYMON $XYMSRV status $host $co $status \"$MSG\"]
    } else {
      exec $XYMON $XYMSRV status $host $co $status \"$MSG\"
    }

  } else {

    # Sending a status with Xymon version 4.3
    # Usage: xymon [--debug] [--proxy=http://ip:port/] RECIPIENT DATA
    # RECIPIENT: IP-address, hostname or URL
    # DATA: Message to send, or "-" to read from stdin
    set MESSAGE "status $host.$co $status $MSG"
    if { $myparams(n) } {
      puts [exec $XYMON $XYMSRV $MESSAGE]
    } else {
      exec $XYMON $XYMSRV $MESSAGE
    }

  }
  set now [ clock click -milliseconds]
  set line [list $host $co $status [expr {($now - $mystart)/1e3}]]
  S add row $line
}

proc send_bb_data { host co MSG } {
#############################################################################
# In here we will take the data contained in MSG and send it to Xymon as a  #
# data update in a single message.  The host and column to which the data   #
# is to apply are given as arguments.  The utility to be used is a local    #
# binary, these have different syntax depending on the platform and         #
# implementation.                                                           #
#############################################################################
  global XYMON XYMSRV XYMONCOMMS OS
  #need to clean dots from host and check XYMSRV, 0.0.0.0 will defer lookup
  #to the XYMON cmd, otherwise need to loop through IP addresses.

  if { $OS == "Windows" } {
    # Sending data with BBWin
    # bbwincmd.exe <bbdisplay>[:<port>] data <hostname> <dataname> <text>
    puts [exec $XYMON $XYMSRV data $host $co \"$MSG\"]

  } else {

    # Sending data with Xymon version 4.3
    # Usage: xymon [--debug] [--proxy=http://ip:port/] RECIPIENT DATA
    # RECIPIENT: IP-address, hostname or URL
    # DATA: Message to send, or "-" to read from stdin
    set MESSAGE "data $host.$co\n$MSG"
    puts [exec $XYMON $XYMSRV $MESSAGE]
  }
}

proc setstatus { attr j mb } {
  set status "green"
  foreach {a} $attr {
    set i [lindex [lindex [V search -nocase -exact column 0 $a] 0] 1]
    if { [string length $i] > 0 } {
      set d [V get cell 1 $i]
      set colour [check_val $j $mb $a $d]
      V set cell 2 $i $colour
      set status [setworstcolour $status $colour]
    } else {
      # should never happen
      set status "yellow"
    }
  }
  return $status
}

proc setworstcolour { worst bbcolor } {
  debug_print "setting worst=$worst or bbcolor=$bbcolor"
  if { [string match $worst "red"] != 1  && [string match $worst "yellow"] != 1 } {
    return $bbcolor
  } elseif { [string match $worst "red" ] != 1 && [string match $bbcolor "yellow"] == 1 } {
    return $bbcolor
  } elseif { [string match $bbcolor "red"] == 1 } {
    return $bbcolor
  } else {
    return $worst
  }
}

proc setbbcolour {pctused warn panic} {
  #debug_print "checking pctused=$pctused < panic=$panic or warn=$warn"
  # these are expected to be numeric values
  if {$pctused >= $panic} {
    return "red"
  } elseif {$pctused >= $warn} {
    return "yellow"
  }
  return "green"
}

proc setbbcolour2 {pctused warn panic} {
  #debug_print "checking pctused=$pctused > panic=$panic or warn=$warn"
  # these are expected to be numeric values
  if {$pctused <= $warn} {
    if {$pctused <= $panic} {
      return "red"
    } else {
      return "yellow"
    }
  }
  return "green"
}

proc setbbcolour3 {pctused warn panic} {
  #debug_print "checking pctused=$pctused != panic=$panic or warn=$warn"
  # these are expected to be string values
  if {[string match -nocase $panic $pctused]} {
    return "red"
  } elseif {[string match -nocase $warn $pctused]} {
    return "yellow"
  }
  return "green"
}

proc setbbcolour4 {pctused warn panic} {
  #debug_print "checking pctused=$pctused = panic=$panic or warn=$warn"
  # these are expected to be string values
  # this test is pass or fail
  if {[string match -nocase $panic $pctused]} {
    return "green"
  } elseif {[string match -nocase $warn $pctused]} {
    return "green"
  }
  return "red"
}

proc jmx_connect_proc_1 {args} {
global status
  set cmd ""
  foreach arg $args {
    set cmd "$cmd$arg "
  }
  puts "command is jmx_connect $cmd"
  set status [eval jmx_connect $cmd]
}

proc jmx_connect_proc {args} {
  global state available connected failed
  set cmd jmx_connect
  foreach arg $args {
    set cmd "$cmd $arg"
  }
  #puts $cmd
  # this will timeout eventually, might be able to modify the underlying
  # code for jmx_connect to do this (implement -t option ? )
  #http://weblogs.java.net/blog/emcmanus/archive/2007/05/making_a_jmx_co_1.html

  #2# set myafter [after 5000 set status 1]
  #2# puts "waiting 5000 on $myafter"
  #2# after 1000 set xxx [catch {jmx_connect_proc_1 $args} result]
  #2# puts "waiting 5000 for $myafter and 250 for jmx_connect_proc_1 $args"
  #2# vwait status
  #2# after cancel $myafter

  set status [catch {eval $cmd} result]

  if { $status == 0 } {
    # The command succeeded, and wrote nothing to stderr.
    # $result contains what it wrote to stdout
    set state $connected
    return 1
  } else {
    set state $failed
    #puts "some error occurred"
    return -1
  }
}

proc remove_pw {text_event} {
  regsub -all -- {jmxremote.port=[0-9]*} $text_event "jmxremote.port=xxxx" text_event
  regsub -all -- {jmxremote.password.file=(.*)} $text_event "jmxremote.password.file=xxxx" text_event
  regsub -all -- {-username\s(.*)} $text_event "-username xxxx" text_event
  regsub -all -- {-password\s(.*)} $text_event "-password xxxx" text_event
  regsub -all -- {-userconfigfile\s(.*)} $text_event "-userconfigfile xxxx" text_event
  regsub -all -- {-userkeyfile\s(.*)} $text_event "-userkeyfile xxxx" text_event
  return $text_event
}

proc convtime {tm} {
  set d [expr int($tm / 86400)]
  set h [expr int(($tm % 86400) / 3600)]
  set m [expr int(($tm % 3600) / 60)]
  set s [expr int($tm % 60)]
  return [expr {$d > 0 ? [format "%dd %02d:%02d:%02d" $d $h $m $s]:[format "%02d:%02d:%02d" $h $m $s]}]
}

proc conv_to_sec {args} {
  return 0
}

proc get_host_avail { h co } {
  global available disconnected
  # check if the host h can be contacted
  set fields [ split [query_bb $h $co] "|"]
  #cscsvradc008.spiderlink.local|conn|green||1299588755|1300651827|1300653627|0|0|20.157.34.26|-1|||N
  #green
  #conn status for cscsvradc008.spiderlink.local at 20:10:26
  #
  if { [string match -nocase [lindex $fields 2] "red"] } {
    return $disconnected
  } else {
    return $available
  }
}

proc set_hostcomma {h} {
  regsub -all -- {\.} $h "," hh
  return $hh
}

proc print_data_full { j c hosts msgs } {
#############################################################################
# In here we will construct an html summary page to be sent to Xymon as a   #
# single message which is returned as a string.  The page will be annotated #
# with the jmxstat host <j> and is destined for column <c>.                 #
#############################################################################
  global Command version S
  upvar #0 begin mystart
  # S:host:column:colour:time

  append result "$c status for $j at [clock format [clock seconds] -format {%H:%M:%S}]\n\n"
  set end [ clock click -milliseconds]

  append result [format "Test Time = %.2f seconds\n" [expr {($end - $mystart)/1e3}]]
  append result [format "Hosts checked = %d\n" $hosts]
  append result [format "Attributes collected = %d\n" $msgs]

  # get a list of colurs reported
  set last ""
  set aco ""
  set hco [format "%-26s" "Host"]
  foreach {co} [lsort [S get column 2]] {
    if { $co != $last } {
      lappend aco $co
      append hco [format " %6s" $co]
    }
    set last $co
  }
  append result [format "\nColours\n"]
  foreach {co} $aco {
    append result [format "%-10s %d\n" $co [llength [S search -nocase -exact column 2 $co]]]
  }

  # get a list of Events reported
  set last ""
  set aev ""
  set hev [format "%-26s" "Host"]
  foreach {ev} [lsort [S get column 1]] {
    if { ($ev != $last) && ( ![regexp -nocase {(error|warning)} $ev] ) } {
      lappend aev $ev
      append hev [format " %10s" $ev]
    }
    set last $ev
  }
  append result [format "\nEvents\n"]
  foreach {ev} $aev {
    append result [format "%-10s %d\n" $ev [llength [S search -nocase -exact column 1 $ev]]]
  }

  # get a list of hosts reported
  set last ""
  set aho ""
  foreach {ho} [lsort [S get column 0]] {
    if { $ho != $last } {
      lappend aho $ho
    }
    set last $ho
  }
  append result [format "\nSummary of Colours by Host\n"]

  append result [format "%s %6s %7s\n" $hco "Total" "Time"]
  foreach {ho} $aho {
    set table ""
    set t 0
    foreach {co} $aco {
      set i 0
      foreach {cell} [S search -nocase -exact column 2 $co] {
        set row [lindex $cell 1]
        if { [string match $ho [S get cell 0 $row] ] } { incr i 1 } 
      }
      set t [expr $t + $i]
      append table [format " %6d" $i]
    }
    if { $t } {
      append table [format " %6d" $t]
      set time ""
      foreach {cell} [S search -nocase -exact column 0 $ho] {
        if { ![string length $time] } {
          set row [lindex $cell 1]
          if { (![string length [S get cell 1 $row]]) && (![string length [S get cell 2 $row]]) } {
            set time [S get cell 3 $row]
          } 
        } 
      }
      if { [string length $time] } {
        append table [format " %7.2f" $time]
      } else {
        append table [format " -"]
      }
      append result [format "%-26s" $ho]
      append result [format "%s\n" $table]
    }
  }

  append result [format "\nSummary of Events by Host\n"]
  append result [format "%s\n" $hev]
  foreach {ho} $aho {
    set table ""
    foreach {ev} $aev {
      set i 0
      foreach {cell} [S search -nocase -exact column 1 $ev] {
        set row [lindex $cell 1]
        if { [string match $ho [S get cell 0 $row] ] } { incr i 1 } 
      }
      if { $i } {
        append table [format " %10d" $i]
      } else {
        append table [format " %10s" "-"]
      }
    }
    if { [string length $table] } {
      append result [format "%-26s" $ho]
      append result [format "%s\n" $table]
    }
  }

  # get a list of Errors reported
  # S:host:error:colour:message
  append result [format "\nErrors\n"]
  set i [llength [S search -nocase -exact column 1 "error"]]
  if { $i } {
    foreach {cell} [S search -nocase -exact column 1 "error"] {
      set row [lindex $cell 1]
      append result [format "%-26s %s\n" [S get cell 0 $row] [S get cell 3 $row]]
    }
  } else {
    append result [format "%s\n" "No General errors"]
    append result [format "%s\n" "No Host errors"]
  }

  # get a list of Warnings reported
  # S:host:warning:colour:message
  append result [format "\nWarnings\n"]
  set i [llength [S search -nocase -exact column 1 "warning"]]
  if { $i } {
    foreach {cell} [S search -nocase -exact column 1 "warning"] {
      set row [lindex $cell 1]
      append result [format "%-26s %s\n" [S get cell 0 $row] [S get cell 3 $row]]
    }
  } else {
    append result [format "%s\n" "No General warnings"]
    append result [format "%s\n" "No Host warnings"]
  }

  append result "\n$Command "
  append result "version $version "
  append result "column $c "
  append result [format "tested in %.2f seconds" [expr {($end - $mystart)/1e3}]]
  return $result
}

proc format_server_body { } {
##########################################################
# In here we will construct the body for the Server page #
##########################################################

  for {set i 0} {$i < [V rows]} {incr i 1} { 
    #set colour [V get cell 2 $i]
    #append body "\t&$colour\t[V get cell 0 $i] : [V get cell 1 $i]\n"

    if { [string length [V get cell 0 $i]] > 0 } {
      switch -- [V get cell 0 $i] {
        StartTime {
          set d [V get cell 1 $i]
          append d "e0"
          append body "\t[V get cell 0 $i] : [clock format [expr int($d/1000)]]\n"
        } 
        Uptime {
          set d [V get cell 1 $i]
          append d "e0"
          append body "\t[V get cell 0 $i] : [convtime [expr int($d/1000)]]\n"
        } 
        default {
          append body "\t[V get cell 0 $i] : [V get cell 1 $i]\n"
        } 
      }
    } else {
      # InputArguments : on seperate lines
      append body "\t                 [V get cell 1 $i]\n"
    }

    if { [string length [V get cell 3 $i]] > 0 } {
      append data "[V get cell 3 $i]:[V get cell 1 $i]\n"
    }
  }
  return $body
}

proc format_threads_body { } {
###########################################################
# In here we will construct the body for the Threads page #
#                                                         #
# Xymon graph definition as follows:                      #
# NCV_Threads="LiveThreads:GAUGE,                         #
#              Peak:GAUGE,                                #
#              DaemonThreads:GAUGE,                       #
#              Total:DERIVE"                              #
#                                                         #
#    Live Threads            18           Peak 20         #
#    Daemon Threads          17  Total Started 305        #
#    <!--                                                 #
#    LiveThreads:18                                       #
#    Peak:20                                              #
#    DaemonThreads:17                                     #
#    Total:305                                            #
#      -->                                                #
###########################################################

  append data "<!--  \n"

  set i [lindex [lindex [V search -nocase -exact column 0 "ThreadCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "Live Threads" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "PeakThreadCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "Peak" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "DaemonThreadCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "Daemon Threads" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "TotalStartedThreadCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "Total Started" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  append data "  -->"
  append body "$data\n"
  return $body
}

proc format_classes_body { } {
###########################################################
# In here we will construct the body for the Classes page #
#                                                         #
# Xymon graph definition as follows:                      #
# NCV_Classes="Classes:GAUGE"                             #
#                                                         #
#    Current loaded        2015     Total unloaded      0 #
#    Total loaded          2015                           #
#    <!--                                                 #
#    Classes:2015                                         #
#      -->                                                #
###########################################################

  append data "<!--  \n"

  set i [lindex [lindex [V search -nocase -exact column 0 "LoadedClassCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "Current loaded" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "UnloadedClassCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "Total unloaded" [V get cell 1 $i]]

  set i [lindex [lindex [V search -nocase -exact column 0 "TotalLoadedClassCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s\n" "Total loaded" [V get cell 1 $i]]

  append data "  -->"
  append body "$data\n"
  return $body
}

proc format_service_body { } {
################################################################
# In here we will construct the body for the Services page     #
# Attributes from MBean Catalina:type=GlobalRequestProcessor.* #
#                                                              #
# Xymon graph definition as follows:                           #
# NCV_Service="bytesReceived:DERIVE,                           #
#              bytesSent:DERIVE,                               #
#              errorCount:DERIVE,                              #
#              requestCount:DERIVE"                            #
#                                                              #
#    bytesReceived       0           bytesSent 0               #
#    maxTime             0      processingTime 0               #
#    requestCount        0          errorCount 0               #
#    <!--                                                      #
#    bytesReceived:0                                           #
#    bytesSent:0                                               #
#    requestCount:0                                            #
#    errorCount:0                                              #
#      -->`                                                    #
################################################################

  append data "<!--  \n"

  set i [lindex [lindex [V search -nocase -exact column 0 "bytesReceived"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "bytesReceived" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "bytesSent"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "bytesSent" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "maxTime"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "maxTime" [V get cell 1 $i]]

  set i [lindex [lindex [V search -nocase -exact column 0 "processingTime"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "processingTime" [V get cell 1 $i]]

  set i [lindex [lindex [V search -nocase -exact column 0 "requestCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "requestCount" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "errorCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "errorCount" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  append data "  -->"
  append body "$data\n"
  return $body
}

proc format_jvm_body { } {
###########################################################
# In here we will construct the body for the JVMHeap page #
# Attributes from MBean java.lang:type=Memory             #
#                                                         #
# Xymon graph definition as follows:                      #
# NCV_JVMHeap="HeapSize:GAUGE,                            #
#              Committed:GAUGE,                           #
#              MaxHeapSize:GAUGE"                         #
#                                                         #
#    Current Heap Size  6019     Committed Memory 7744    #
#    Maximum Heap Size 659264     Initialised Memory 0    #
#    <!--                                                 #
#    HeapSize:6019                                        #
#    Committed:7744                                       #
#    MaxHeapSize:659264                                   #
#      -->                                                #
###########################################################

  append data "<!--  \n"

  set i [lindex [lindex [V search -nocase -exact column 0 "HeapMemoryUsageUsed"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-18s %6s kbytes     " "Current Heap Size" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "HeapMemoryUsageCommitted"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%18s %6s kbytes\n" "Committed Memory" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "HeapMemoryUsageMax"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-18s %6s kbytes     " "Maximum Heap Size" [V get cell 1 $i]]
  append data "[V get cell 3 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "HeapMemoryUsageInit"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%18s %6s kbytes\n" "Initialised Memory" [V get cell 1 $i]]

  append data "  -->"
  append body "$data\n"
  return $body
}

proc format_gc_body { co host } {
##############################################################################
# In here we will construct the body for the GCInfo page                     #
# Attributes from MBean java.lang:type=GarbageCollector,name=*               #
#                                                                            #
# Xymon graph definition as follows:                                         #
#    DS:CollectionCount:DERIVE:600:0:U                                       #
#    DS:CollectionTime:DERIVE:600:0:U                                        #
#                                                                            #
# but note that this data is required in seperate RRD files for each         #
# collector type.  This is achieved with special processing configured on    #
# the data channel on the server via the extra-tests option option to        #
# hobbitd_channel, eg :-                                                     #
#       --extra-tests=GCInfo                                                 #
# The RRD information is therefore submitted to Xymon using the data channel #
# via a seperate "xymon data" message                                        #
#                                                                            #
# Collector(s):                                                              #
#            Copy             Collections=1025   Total Time spent=0.835 secs #
#            MarkSweepCompact Collections=0      Total Time spent=0.000 secs #
#                                                                            #
# Most Recent Collection(s):                                                 #
#            Copy                    0.000 sec       1 sec ago               #
#                                                                            #
#                                         Used       Recovered               #
#            Eden Space            2176 kbytes     2176 kbytes               #
#            Tenured Gen           4919 kbytes        0 kbytes               #
#            Perm Gen             15787 kbytes        0 kbytes               #
#                                                  2110 kbytes               #
#                                                                            #
#            MarkSweepCompact  9197725.694 sec 43 days 17 hrs 56 mins ago    #
#                                                                            #
#                                         Used       Recovered               #
#                                                     0 kbytes               #
#                                                                            #
# 20.130.151.87 data tomcat_fosportal_008.GCInfo                             #
#                                                                            #
# GarbageCollectors: Copy MarkSweepCompact                                   #
# CopyCollectionCount:1025                                                   #
# CopyCollectionTime:835                                                     #
# MarkSweepCompactCollectionCount:0                                          #
# MarkSweepCompactCollectionTime:0                                           #
#                                                                            #
##############################################################################
  append data "\nGarbageCollectors: "
  set body ""
  set gcs ""

  # get a list of Garbage Collectors - deal with spaces
  set last ""
  foreach {gc} [lsort [V get column 0]] {
    if { $gc != $last } {
      regsub -all { } $gc "_" rrd
      lappend gcs $rrd
    }
    set last $gc
  }
  append data "$gcs\n"

  # Announce the collection information for each collector
  # NB some collectors may not have started yet
  #append body [format "\t%s\t%s\t%s\n" "Garbage Collector Name" "Collections" "Time Spent"]
  foreach {rrd} $gcs {
    regsub -all {_} $rrd " " gc
    if { [string length $body] == 0 } {
      #append body [format "%50s\n" "Garbage Collection"]
      append body [format "%s\n" "Collector(s):"]
    }
    # select the rows applicable to this collector
    set rows ""
    set gccount ""
    foreach {c} [V search -nocase -exact column 0 $gc] {
      lappend rows [lindex $c 1]
    }
    foreach {i} $rows {
      debug_print "[V get cell 3 $i]:[V get cell 1 $i]"
      if { [V get cell 3 $i] == "CollectionCount"} {
        set gccount [V get cell 1 $i]
        append data $rrd [V get cell 3 $i] ":" $gccount "\n"
      }
      if { [V get cell 3 $i] == "CollectionTime"} {
        set gctime [V get cell 1 $i]
        append data $rrd [V get cell 3 $i] ":" $gctime "\n"
      }
    }
    if { ([string length $gccount] > 0) && ($gccount > 0) } {
      append body [format "\t%-20s" $gc]
      append body [format "Collections=%-7s" $gccount]
      append gctime "e0"
      append body [format "Total Time spent=%7.3f seconds" [expr $gctime/1000]]
      append body "\n"
    } elseif { ([string length $gctime] > 0) && ($gctime > 0) } {
      append body [format "&yellow\t%-20s" $gc]
      append body [format "Collections=%-7s" $gccount]
      append gctime "e0"
      append body [format "Total Time spent=%7.3f seconds" [expr $gctime/1000]]
      append body "\n"
    }
  }
  append body "\n"

  # Announce the last collection details for each collector
  # NB some collectors may not have started yet
  append body [format "%s\n" "Last Collection(s):"]
  foreach {rrd} $gcs {
    # select the rows applicable to this collector
    regsub -all {_} $rrd " " gc
    set rows ""
    set gccount ""
    set pools ""
    set allpools ""
    foreach {c} [V search -nocase -exact column 0 $gc] {
      lappend rows [lindex $c 1]
    }
    foreach {i} $rows {
      # get a list of Memory Pool Names active for this collector
      # NB all pools are recorded, collection only manipulates these pools
      set last ""
      if { [string match "*BeforeGc.*.used" [V get cell 3 $i] ] } {
        regexp {^memoryUsageBeforeGc.(.*).used$} [V get cell 3 $i] -> p
        if { $p != $last } {
          lappend allpools $p
        }
        set last $allpools
      } 
      # ... but collection only manipulates these pools
      set last ""
      if { [V get cell 3 $i] == "MemoryPoolNames"} {
        if { [V get cell 1 $i] != $last } {
          lappend pools [V get cell 1 $i]
        }
        set last $pools
      }
      if { [V get cell 3 $i] == "CollectionCount"} {
        set gccount [V get cell 1 $i]
      }
    }
    if { ([string length $gccount] > 0) && ($gccount > 0) } {
      foreach {i} $rows {
        if { [V get cell 3 $i] == "startTime"} {
          set start [V get cell 1 $i]
        }
        if { [V get cell 3 $i] == "duration"} {
          set duration [V get cell 1 $i]
          set lamp [V get cell 2 $i]
        }
      }
      append duration "e0"
      append body [format "&%s\t" $lamp]
      append body [format "%s" $gc]
      append body [format " ran at %s" [clock format [expr int($start)]]]
      append body [format " and completed in %.3f seconds" [expr $duration/1000]]
      #append body [format " %-20s" [clock format $start]]
      append body "\n"

      append body [format "\t"]
      append body [format "%20s" ""]
      append body [format " %16s" "Used"]
      append body [format " %16s" "Recovered"]
      append body "\n"
      foreach {p} $allpools {
        regsub -all -- {(\[|\])} $p {\\&} pattern
        foreach {i} $rows {
          #Copy:16171032:purple:memoryUsageBeforeGc.Perm Gen.used
          if { [string match "*BeforeGc.$pattern.used" [V get cell 3 $i] ] } {
            set before [V get cell 1 $i]
            append before "e0"
          }
          if { [string match "*AfterGc.$pattern.used" [V get cell 3 $i] ] } {
            set used [V get cell 1 $i]
            append used "e0"
          }
        }
        append body [format "\t"]
        append body [format "%20s" $p]
        append body [format " %9s kbytes" [expr $used/1024]]
        if { [lsearch -exact $pools $p] == -1 } {
          append body [format " %16s" "N/A"]
        } else {
          append body [format " %9s kbytes" [expr ($before-$used)/1024]]
        }
        append body "\n"
      }
    }
  }
  send_bb_data $host $co $data
  return $body
}

proc format_threadpool_body { } {
################################################################
# In here we will construct the body for the Threadpool page   #
# Attributes from MBean Catalina:type=ThreadPool,name=*        #
#                                                              #
# Xymon graph definition as follows:                           #
# NCV_ThreadPool="currentThreadCount:GAUGE,                    #
#                 currentThreadsBusy:GAUGE"                    #
#                                                              #
#    currentThreadCount  4      maxSpareThreads  50            #
#    currentThreadsBusy  1      minSpareThreads  4             #
#    maxThreads          200                                   #
#    <!--                                                      #
#    currentThreadCount: 4                                     #
#    currentThreadsBusy: 1                                     #
#      -->`                                                    #
################################################################

  append data "<!--  \n"

  set i [lindex [lindex [V search -nocase -exact column 0 "currentThreadCount"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "currentThreadCount" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "maxSpareThreads"] 0] 1]
  if { [string length $i] > 0 } {
    append body [format "\t&%-6s\t" [V get cell 2 $i]]
    append body [format "%15s %s\n" "maxSpareThreads" [V get cell 1 $i]]
  } else {
    append body [format "\t\t%s\n" "maxSpareThreads not exposed"]
  }

  set i [lindex [lindex [V search -nocase -exact column 0 "currentThreadsBusy"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "currentThreadsBusy" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "minSpareThreads"] 0] 1]
  if { [string length $i] > 0 } {
    append body [format "\t&%-6s\t" [V get cell 2 $i]]
    append body [format "%15s %s\n" "minSpareThreads" [V get cell 1 $i]]
  } else {
    append body [format "\t\t%s\n" "minSpareThreads not exposed"]
  }

  set i [lindex [lindex [V search -nocase -exact column 0 "maxThreads"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %8s\n" "maxThreads" [V get cell 1 $i]]

  append data "  -->"
  append body "$data\n"
  return $body
}

proc format_webapp_body { } {
#####################################################################
# In here we will construct the body for the WebApp page            #
# Attributes from MBean Catalina:type=Manager,path=*,host=localhost #
#                                                                   #
# Xymon graph definition as follows:                                #
# NCV_WebApp="activeSessions:GAUGE,                                 #
#             sessionCounter:GAUGE,                                 #
#             maxActive:GAUGE,                                      #
#             sessionMaxAliveTime:GAUGE,                            #
#             sessionAverageAlive:GAUGE,                            #
#             rejectedSessions:GAUGE,                               #
#             expiredSessions:GAUGE"                                #
#                                                                   #
#    activeSessions: 0           expiredSessions: 18                #
#    maxActive: 0                sessionCounter: 18                 #
#    sessionMaxAliveTime: 250441 sessionAverageAliveTime: 247425    #
#    rejectedSessions: 0                                            #
#    <!--                                                           #
#    activeSessions: 0                                              #
#    expiredSessions: 18                                            #
#    maxActive: 0                                                   #
#    sessionCounter: 18                                             #
#    sessionMaxAliveTime: 250441                                    #
#    sessionAverageAlive: 247425                                    #
#    rejectedSessions: 0                                            #
#      -->`                                                         #
#####################################################################

  append data "<!--  \n"

  set i [lindex [lindex [V search -nocase -exact column 0 "activeSessions"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "activeSessions" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "expiredSessions"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "expiredSessions" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "maxActive"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "maxActive" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "sessionCounter"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "sessionCounter" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "sessionMaxAliveTime"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s     " "sessionMaxAliveTime" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "sessionAverageAliveTime"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%15s %s\n" "sessionAverageAliveTime" [V get cell 1 $i]]
  #append data "[V get cell 0 $i]:[V get cell 1 $i]\n"
  append data "sessionAverageAlive:[V get cell 1 $i]\n"

  set i [lindex [lindex [V search -nocase -exact column 0 "rejectedSessions"] 0] 1]
  append body [format "\t&%-6s\t" [V get cell 2 $i]]
  append body [format "%-15s %5s\n" "rejectedSessions" [V get cell 1 $i]]
  append data "[V get cell 0 $i]:[V get cell 1 $i]\n"

  append data "  -->"
  append body "$data\n"
  return $body
}

proc format_default_body { } {
#############################################################
# In here we will construct the body for the default mbean. #
# This page layout has no special formatting, we just dump  #
# all the data in the table - note CSV module can read this #
# so we dont send the data seperately.                      #
#############################################################

  global S warn
  #append data "<!--  \n"
  set body ""
  for {set i 0} {$i < [V rows]} {incr i 1} { 
    set colour [V get cell 2 $i]
    append body "\t&$colour\t[V get cell 0 $i] : [V get cell 1 $i]\n"
    if { [string length [V get cell 3 $i]] > 0 } {
      append data "[V get cell 3 $i]:[V get cell 1 $i]\n"
    }
  }
  if { [string length $body] == 0 } {
    append body "No results returned - see warnings page"
  }
  #append data "  -->"
  #append body "$data\n"

  return $body
}

proc fetch_message { j mb co } {
#############################################################################
# In here we will construct an html page to be sent to Xymon as a single    #
# message which is returned as a string.  The page will be simply derived   #
# by fetching the current message for column <co> of the Java instance <j>. #
# Some mbeans need to be reported differently.                              #
#############################################################################

  set lines [ split [query_bb $j $co] "\n"]
  set i 0
  set indata 0

  #tomcat_fosportal_008|Threads|green||1299588755|1300651827|1300653627|0|0|20.157.34.26|-1|||N
  #green
  #Threads status for tomcat_fosportal_008 at 20:10:26
  #
  #Attributes from MBean java.lang:type=Threading
  #
  #        &green  Live Threads       29           &green             Peak 29
  #        &green  Daemon Threads     28           &green    Total Started 5343
  #<!--
  #LiveThreads:29
  #Peak:29
  #DaemonThreads:28
  #Total:5343
  #  -->
  #
  #jmxstat version 0.0.5 column Threads tested in 0.16 seconds

  foreach line $lines {
    debug_print "Line $i/[llength $lines]: $line"
    regsub -all {(\<!--.*--\>)} $line {} line
    if { $i >= 2 } {
      if { [regsub -all -- {\<!--} $line {} line] } { set indata 1 }
      if { ($indata) && ([regsub -all -- {--\>} $line {} line]) } {
        set indata 0
      } elseif { !$indata } {
        regsub -all -nocase {&(red|yellow|green)} $line "\\&clear" line
        append result "$line\n"
      }
    }
    incr i
  }
  #regsub -all {(\<\!--.*--\>)} $result {} result
  return $result
}

proc wrap_message { j mb co } {
#############################################################################
# In here we will construct an html page to be sent to Xymon as a single    #
# message which is returned as a string.  The page will be annotated with   #
# the Java instance <j> and is destined for column <co>.  Some mbeans need  #
# to be reported differently.  The data for the page comes from the         #
# contents of the global array <V>                                          #
#############################################################################

  global Command version
  upvar #0 beanstart mystart

  ##########################################
  # create a standard header for the page
  ##########################################
  #puts "wrap_message: mb=$mb"
  append result "$co status for $j at [clock format [clock seconds] -format {%H:%M:%S}]\n\n"
  if { ![string match "Uptime" $mb] } {
    append result "Attributes from MBean $mb\n\n"
  }
  #put the values in html comments like <!--  PinSQLArea:99.80  PinTblProc:99.02  PinBody:99.97  PinTrigger:99.94  HitSQLArea:95.58  HitTblProc:93.14  HitBody:99.95  HitTrigger:99.56  BlBuffHit:98.57  RowCache:97.26  -->  
  append data "<!--  \n"

  ###############################################################
  # format of the page body depends on the mbean being reported
  ###############################################################
  debug_print "there are [V rows] rows "
  debug_print "and [V columns] columns at the moment\n"
  switch -regexp -- $mb {

    (Uptime) {
      # special formatting for the fake Uptime page
      set start [V get cell 1 0]
      append start "e0"
      append result "Server $j was restarted at [clock format [expr int($start/1000)]]\n"
      set uptime [V get cell 1 1]
      append uptime "e0"
      append result "System Uptime is [convtime [expr int($uptime/1000)]]\n"
    }

    java.lang:type=Runtime { append result [format_server_body] }
    java.lang:type=Threading { append result [format_threads_body] }
    java.lang:type=ClassLoading { append result [format_classes_body] }
    Catalina:type=GlobalRequestProcessor { append result [format_service_body] }
    java.lang:type=Memory { append result [format_jvm_body] }
    Catalina:type=ThreadPool,name=.* { append result [format_threadpool_body] }
    Catalina:type=Manager,path=.*,host=localhost { append result [format_webapp_body] }
    java.lang:type=(GarbageCollector) { append result [format_gc_body $co $j ] }
    default { append result [format_default_body] }
  }

  ##########################################
  # create a standard footer for the page
  ##########################################
  set end [ clock click -milliseconds]
  append result "\n$Command "
  append result "version $version "
  append result "column $co "
  append result [format "tested in %.2f seconds" [expr {($end - $mystart)/1e3}]]
  return $result
}

proc check_val { j mb attr current } {
###########################################################################
# In here we will look up in the INI file the attribute given in <attr>   #
# for mbean <mb> either in the section for the Java instance <j>, or the  #
# default section if no Java instance is set.  We check any defined alert #
# specification against the supplied value <current>.  The Xymon status   #
# colour will be returned.                                                #
###########################################################################
  global myfh
  set result "green"
  set spec ""
  set operator "<"
  set limit ""
  debug_print "checking JVM $j or MBean $mb for attr $attr"

  # Query built-in defaults - specifically in case Uptime thresholds
  # are missing from the ini file
  set spec [init_status $attr]

  # this aborts if section or key cant be found - configuration error
  # should catch this really
  if { [ini::exists $myfh $mb $attr] } {
    set spec [string trim [ini::value $myfh $mb $attr]]
    debug_print "default spec for $attr is $spec"
  }
  if { [ini::exists $myfh $j $attr] } {
    set spec [string trim [ini::value $myfh $j $attr]]
    debug_print "instance spec for $attr in $j is $spec"
  }

  if { [string length $spec] == 0 } {
    # We have no test to perform
    return "green"
  }

  # We have some sort of test to perform
  # 80;90;<;HeapMemoryUsageInit
  # Split into fields on semi-colons
  set fields [split $spec ";"]

  switch [llength $fields] {
    4 {
        mylassign $fields warn fail operator limit
        if { [string length $limit] > 0 } {
          # check if we have limit available in the matrix yet
          set a [lindex [lindex [V search -nocase -exact column 0 $limit] 0] 1]
          if { [string length $a] > 0 } {
            set warn [ expr [V get cell 1 $a]*$warn/100]
            set fail [ expr [V get cell 1 $a]*$fail/100]
            debug_print "reset warn to $warn and fail to $fail"
          } else {
            debug_print "check_val:required resource ($limit) not available yet"
            return "yellow"
          }
        }
      }
    3 {
        mylassign $fields warn fail operator
        if { [string length $operator] == 0 } { set operator "<" }
      }
    2 {
        mylassign $fields warn fail
      }
    default {
        puts "check_val:invalid alert spec"
        return "yellow"
      }
  }
  debug_print "warn=$warn, fail=$fail, operator=$operator, limit=$limit"

  switch -- $operator {
    < {
        debug_print "checking current=$current less than fail=$fail"
        set result [setbbcolour $current $warn $fail]
      }
    > {
        debug_print "checking current=$current greater than fail=$fail"
        set result [setbbcolour2 $current $warn $fail]
      }
    {!} {
        debug_print "checking current=$current not equal to $fail"
        set result [setbbcolour3 $current $warn $fail]
      }
    {=} {
        debug_print "checking current=$current equals $fail"
        set result [setbbcolour4 $current $warn $fail]
      }
    default {
        mydie "unknown operator ($operator)"
      }
  }

  return $result
}

proc jmx_lookup { j mb co attr } {
#############################################################################
# In here we will look up the attributes given in <attr> for mbean <mb>     #
# and for each attribute, check for alert condition and build data for the  #
# page under column <co> in the global array <V>.  Lastly the page will be  #
# constructed in proc wrap_page and sent to Xymon using proc send_bb.       #
# Some attributes and some mbeans need to be treated specially              #
# we return the number of attributes we have captured.                      #
#############################################################################
  global myfh XYMON state connected MBEAN ATTROP
  global P
  global S warn
  ::struct::matrix  V = P

  debug_print "Processing $mb for column $co"

  set i 0
  set attrok 0
  # mostly we defer setting individual colours till after all attributes
  # fetched since there might be alerts based on values not yet processed
  # but some mbeans we still process as we go
  if { $state != $connected } {
    set status "clear"
    send_bb $j $co $status [fetch_message $j $mb $co]
  } else {
    set status "green"
    set colour "purple"
    switch -regexp -- $mb {

      (Uptime) {
        # uptime processing
        debug_print "Fetching attributes for fake uptime"
        set rmb "java.lang:type=Runtime"

        foreach {a} $attr {
          set d [jmx_get -m $rmb $a]
          if { [string match "Uptime" $a] == 1 } {
            set m $d
            append m "e0"
            set colour [check_val $j $rmb $a [expr int($m/1000/60)]]
            #set line [list $a $d $colour ""]
            #set status [setworstcolour $status $colour]
          } else {
            set colour "green"
            #set line [list $a $d "green" ""]
            #set status [setworstcolour $status "green"]
          }
          set line [list $a $d $colour ""]
          V add row $line
          set status [setworstcolour $status $colour]
          set i [expr $i + 1]
        }
      }

      java.lang:type=(Runtime) {
        # runtime mbean processing
        # legacy support, the ini files were not previously honoured and
        # may contain junk.  What we need here is
        # VmName VmVersion VmVendor StartTime ClassPath LibraryPath BootClassPath
        # we should probably weed out 
        #   StartTime Uptime
        # which are covered under Uptime column
        set attr "VmName VmVersion VmVendor StartTime Uptime ClassPath LibraryPath BootClassPath InputArguments"
        foreach {a} $attr {
          set d [jmx_get -m $mb $a]
          switch -regexp -- $a {
            ^InputArguments$ {
              # InputArguments returns a java String array
              set java_obj [jmx_get -n -m $mb $a]
              # Need to use Java/tcl to convert to Tcl list of strings
              set args [split [$java_obj getrange]]
              set A $a
              foreach {d} $args {
                set line [list $A [remove_pw $d] "green" ""]
                V add row $line
                set i [expr $i + 1]
                set A ""
              }
            }
            Uptime {
              set colour "green"
              set line [list $a $d $colour ""]
              V add row $line
              set status [setworstcolour $status $colour]
              set i [expr $i + 1]
            }
            default {
              set colour [check_val $j $mb $a $d]
              set line [list $a $d $colour ""]
              V add row $line
              set status [setworstcolour $status $colour]
              set i [expr $i + 1]
            }
          }
        }
      }

      java.lang:type=(Memory) {
        # special processing for Memory mbean
        debug_print "Fetching attributes for Java Heap"
        # maybe should swap cattr with xymon attribute name?
        set cattr ""
        foreach {a} $attr {
          set d [jmx_get -m $mb $a]
          #debug_print "$a $d"
          switch -regexp -- $a {
            ^HeapMemoryUsage$ {
              regexp {^.*committed=([0-9]*),.*init=([0-9]*),.*max=([0-9]*),.*used=([0-9]*)} $d -> committed init max used
              debug_print "$d"

              append composite $a "Committed"
                append committed "e0"
                set committed [expr int($committed/1024)]
                debug_print "got $composite=$committed"
                set line [list $composite $committed $colour "Committed"]
                V add row $line
                set i [expr $i + 1]
                lappend cattr $composite
                set composite ""

              append composite $a "Init"
                append init "e0"
                set init [expr int($init/1024)]
                debug_print "got $composite=$init"
                set line [list $composite $init $colour "Initialised"]
                V add row $line
                set i [expr $i + 1]
                lappend cattr $composite
                set composite ""

              append composite $a "Max"
                append max "e0"
                set max [expr int($max/1024)]
                debug_print "got $composite=$max"
                set line [list $composite $max $colour "MaxHeapSize"]
                V add row $line
                set i [expr $i + 1]
                lappend cattr $composite
                set composite ""

              append composite $a "Used"
                append used "e0"
                set used [expr int($used/1024)]
                debug_print "got $composite=$used"
                set line [list $composite $used $colour "HeapSize"]
                V add row $line
                set i [expr $i + 1]
                lappend cattr $composite
                set composite ""
            }
            NonHeapMemoryUsage {
            }
            default {
              set line [list $a $d $colour ""]
              V add row $line
              set i [expr $i + 1]
              lappend cattr $a
            }
          }
        }
        set status [ setstatus $cattr $j $mb ] 
      }

      java.lang:type=GarbageCollector.* {
        # There are many collectors that can be configured and these
        # can change dynamically.  Record what was requested, but
        # report what collectors are actually current 
        set omb $mb
        set boottime [jmx_get -m "java.lang:type=Runtime" "StartTime"]
        set uptime [jmx_get -m "java.lang:type=Runtime" "Uptime"]
        append boottime "e0"
        append uptime "e0"
        foreach mb [jmx_list java.lang:.*type=GarbageCollector,name=.*]  {
          # special processing for GC Info mbean
          # looping on java.lang:type=GarbageCollector,name=PS_Scavenge
          # Legacy support, remove the underbar
          # regsub ?switches? exp string subSpec varName
          regsub -all {_} $mb " " mb
          debug_print "Fetching attributes $attr for mbean $mb"

          set gc [jmx_get -m $mb "Name"]

          # ignoring attrs for the moment
          if { [catch {jmx_get -m $mb LastGcInfo} LastGcInfo] } {
            #mydie "Problem in GCInfo"
            set a "CollectionTime"
            set d [jmx_get -m $mb $a]
            debug_print "$a $d"
            set line [list $gc $d "yellow" $a]
            V add row $line
            incr i
            set a "CollectionCount"
            set cc 0
            debug_print "$a $cc"
            set line [list $gc $cc "yellow" $a]
            V add row $line
            incr i
          } else {
            regsub -all {,} $LastGcInfo "\n" TextLastGcInfo
            debug_print "found LastGcInfo=$TextLastGcInfo"
            foreach {a} {CollectionTime CollectionCount} {
              set d [jmx_get -m $mb $a]
              debug_print "$a $d"
              set line [list $gc $d "green" $a]
              V add row $line
              incr i
            }
            set cc $d
          }

          # MemoryPoolNames returns a java String array
          set a "MemoryPoolNames"
          set java_obj [jmx_get -n -m $mb $a]
          # Need to use Java/tcl to convert to Tcl list of strings
          set args [split [$java_obj getrange]]
          regsub -all {\\} $args "" args
          set A $a
          foreach {d} $args {
            debug_print "$a $d"
            set line [list $gc $d "green" $a]
            V add row $line
            incr i
            set A ""
          }
          if { $cc > 0 } {

            # now the composite ones ...
            if { [regexp {^.*GcThreadCount=([0-9]*).*endTime=([0-9]*).*id=([0-9]*).*startTime=([0-9]*).*$} $LastGcInfo -> GcThreadCount endTime id startTime] } {
              append endTime "e0"
              append startTime "e0"
              set duration [expr int($endTime - $startTime)]
            } else {
              set GcThreadCount -1
              set endTime -1
              set id -1
              set startTime -1
            }
            #debug_print "GcThreadCount=$GcThreadCount"
            #debug_print "duration=$duration"
            #debug_print "endTime=$endTime"
            #debug_print "id=$id"
            #debug_print "startTime=$startTime"
            foreach {a} {GcThreadCount duration endTime id startTime} {
              switch -regexp -- $a {
                GcThreadCount { set d $GcThreadCount }
                duration { set d $duration }
                endTime { set d [expr int($boottime/1000 + $endTime/1000)] }
                id { set d $id }
                startTime { set d [expr int($boottime/1000 + $startTime/1000)] }
              }
              debug_print "$a $d"
              set gca "LastGcInfo$gc$a"
              set expiredalert [expr 30 * 60]
              switch -regexp -- $a {
                duration { 
                  if { [expr [clock seconds] - [expr int($boottime/1000 + $endTime/1000)]] > $expiredalert } {
                    set colour "green"
                  } else {
                    set colour [check_val $j $mb $gca $d]
                  }
                }
                default {
                  set colour [check_val $j $mb $gca $d]
                }
              }
              set line [list $gc $d $colour $a]
              V add row $line
              set status [setworstcolour $status $colour]
              incr i
            }

            # split out the strings containing the data we need
            regexp {^.*GcInfoCompositeType(.*GcThreadCount).*memoryUsageAfterGc(.*memoryUsageBeforeGc)(.*startTime).*$} $LastGcInfo -> Structure memoryUsageAfterGc memoryUsageBeforeGc startTime
            #debug_print "here is the structure $Structure"
            #debug_print "here is the memoryUsageAfterGc $memoryUsageAfterGc"
            #regexp {^(.*contents=.key=)?(.*)$} $memoryUsageAfterGc -> crap

            # sample 1 -  memoryUsageBeforeGc
            set args [::textutil::split::splitx $memoryUsageBeforeGc {contents=.key=}]
            set sample "memoryUsageBeforeGc"
            set k 0
            foreach {a} $args {
              if { $k == 0 } {
                set names [::textutil::split::splitx $a {itemName=}]
                foreach {n} $names {
                  #debug_print "name=$n"
                }
              } else {
                #debug_print $a
                regexp {^(.*), value=javax.*committed=([0-9]*).*init=([0-9]*).*max=([0-9]*).*used=([0-9]*).*$} $a -> area committed init max used
                #debug_print "area $area, committed=$committed, init=$init, max=$max, used=$used"

                set a $sample
                append a "." $area ".init"
                set line [list $gc $init "green" $a]
                V add row $line
                incr i

                set a $sample
                append a "." $area ".max"
                set line [list $gc $max "green" $a]
                V add row $line
                incr i

                set a $sample
                append a "." $area ".used"
                set line [list $gc $used "green" $a]
                V add row $line
                incr i

                set a $sample
                append a "." $area ".committed"
                set line [list $gc $committed "green" $a]
                V add row $line
                incr i

              }
              incr k
            }

            # sample 2 -  memoryUsageAfterGc
            set args [::textutil::split::splitx $memoryUsageAfterGc {contents=.key=}]
            set sample "memoryUsageAfterGc"
            set k 0
            foreach {a} $args {
              if { $k == 0 } {
                set names [::textutil::split::splitx $a {itemName=}]
                foreach {n} $names {
                  #debug_print "name=$n"
                }
              } else {
                #debug_print $a
                regexp {^(.*), value=javax.*committed=([0-9]*).*init=([0-9]*).*max=([0-9]*).*used=([0-9]*).*$} $a -> area committed init max used
                #debug_print "area $area, committed=$committed, init=$init, max=$max, used=$used"

                set a $sample
                append a "." $area ".init"
                set line [list $gc $init "green" $a]
                V add row $line
                incr i

                set a $sample
                append a "." $area ".max"
                set line [list $gc $max "green" $a]
                V add row $line
                incr i

                set a $sample
                append a "." $area ".used"
                set line [list $gc $used "green" $a]
                V add row $line
                incr i

                set a $sample
                append a "." $area ".committed"
                set line [list $gc $committed "green" $a]
                V add row $line
                incr i

              }
              incr k
            }
          }
        }
      }

      java.lang:type=(ClassLoading|Threading) {
        # special mbean processing - load page with data
        foreach {a} $attr {
          #set d [jmx_get -m $mb $a]
          switch -regexp -- $a {
            ^(UnloadedClassCount|CurrentThreadCpuTime|CurrentThreadUserTime)$ {
              set A ""
            }
            ^LoadedClassCount$ {
              set A "Classes"
            }
            ^TotalLoadedClassCount$ {
              set A "TotalClasses"
            }
            TotalStartedThreadCount {
              set A "Total"
            }
            DaemonThreadCount {
              set A "DaemonThreads"
            }
            PeakThreadCount {
              set A "Peak"
            }
            ThreadCount {
              set A "LiveThreads"
            }
            default {
              set A $a
            }
          }
          if { [catch {jmx_get -m $mb $a} d] } {
            debug_print "could not query attribute $a from mbean $mb\n$d\n"
            set line [list $j "warning" "" ]
            lappend line "could not query attribute $a from mbean $mb"
            S add row $line
            incr warn
            set line [list $a $d "yellow" $A]
          } else {
            debug_print "data for $a returned was $d"
            set line [list $a $d $colour $A]
            incr attrok
          }
          V add row $line
          set i [expr $i + 1]
        }
        #set status [ setstatus $attr $j $mb ]
        if { $i == $attrok } {
          set status [ setstatus $attr $j $mb ]
        } else {
          set status "yellow"
        }
      }

      default {
        # default mbean processing
        foreach {a} $attr {
          debug_print "Fetching attribute $a from mbean $mb"

          # 0.1.7 jcl objects If you specify -n, then the return will be a
          # Java/Tcl java object reference.  See the Java/Tcl documentation
          # on the internet for more details.
          set noconvert ""
          if { [ini::exists $myfh $mb,attr=$a noconvert] } {
            set noconvert [ini::value $myfh $mb,attr=$a noconvert]
          } elseif { [ini::exists $myfh $mb noconvert] } {
            set noconvert [ini::value $myfh $mb noconvert]
          }

          set postexec ""
          if { [ini::exists $myfh $mb,attr=$a post-exec] } {
            set postexec [ini::value $myfh $mb,attr=$a post-exec]
          } elseif { [ini::exists $myfh $mb post-exec] } {
            set postexec [ini::value $myfh $mb post-exec]
          }

          set preexec ""
          if { [ini::exists $myfh $mb,attr=$a pre-exec] } {
            set preexec [ini::value $myfh $mb,attr=$a pre-exec]
          } elseif { [ini::exists $myfh $mb pre-exec] } {
            set preexec [ini::value $myfh $mb pre-exec]
          }

          if { [string length $preexec] > 0 } {
            eval $preexec
          }

          set MBEAN $mb
          set ATTROP $a
          if { [ catch [concat jmx_get $noconvert] d] } {
            debug_print "could not query attribute $a from mbean $mb\n$d\n"
            set line [list $j "warning" "" ]
            lappend line "could not query attribute $a from mbean $mb"
            S add row $line
            incr warn
            set line [list $a $d "yellow" ""]
          } else {

            if { [string length $postexec] > 0 } {
              if { [string length $noconvert] > 0 } {
                debug_print "original data for $a returned a JCL pointer"
              } else {
                debug_print "original data for $a returned was ->|$d|<-"
              }
              # postexec is expected to manipute $d and create a new assignment
              eval $postexec
              debug_print "transformed data for $a returned was ->|$d|<-"
            } else {
              debug_print "data for $a returned was ->|$d|<-"
            }

            set line [list $a $d $colour ""]
            incr attrok
          }
          V add row $line
          set i [expr $i + 1]
        }
        if { $i == $attrok } {
          set status [ setstatus $attr $j $mb ]
        } else {
          set status "yellow"
        }
      }
    }
    send_bb $j $co $status [wrap_message $j $mb $co]
  }
  V destroy
  return $i
}

###############################################################################
# MAIN
###############################################################################
global state debug OS
set debug 0
set available 1
set connected 2
set disconnected 0
set timeoutcmdexec -1
set failed -2
set OS [lindex $tcl_platform(os) 0]

set begin [clock click -milliseconds]
::struct::matrix  P
# attribute name as known within JMX
P add column "attr"
# current value to be reported
P add column "value"
# current status to be reported
P add column "alert"
# sample name in RRD DB
P add column "rrd"

::struct::matrix  S
S add column "host"
S add column "column"
S add column "colour"
S add column "seconds"

# What we put in hosts to trigger this test
set HTAG "jmxstat"

set XYMONCOMMS "xymon"
set XYMONSEARCH "xymongrep"
# inherit this from the environment or startup profile if its there
if { [info exists ::env(XYMONHOME) ] } {
  set XYMONHOME $::env(XYMONHOME)
} elseif { [info exists ::env(BBHOME) ] } {
  set XYMONHOME $::env(BBHOME)
  set XYMONCOMMS "bb"
  set XYMONSEARCH "bbhostgrep"
} else {
  # assume we are running from the ext directory
  set XYMONHOME [file dirname [pwd]]
}
#puts $XYMONHOME

if {[info exists ::env(XYMSRV)]} {
  set XYMSRV $::env(XYMSRV)
} elseif {[info exists ::env(BBDISP)]} {
  set XYMSRV $::env(BBDISP)
  set XYMONCOMMS "bb"
  set XYMONSEARCH "bbhostgrep"
} else {
  set XYMSRV "0.0.0.0"
}
#puts $XYMSRV

#parray ::env
if { [info exists ::env(XYMON) ] } {
  set XYMON $::env(XYMON)
} elseif { [info exists ::env(BB) ] } {
  set XYMON $::env(BB)
  set XYMONCOMMS "bb"
  set XYMONSEARCH "bbhostgrep"
} else {
  if { $OS == "Windows" } {
    set XYMON [ file join $XYMONHOME bin BBWinCmd.exe]
  } else {
    set XYMON [ file join $XYMONHOME bin $XYMONCOMMS ]
  }
}
# just remove it then it doesnt matter if we test untagged
if { [info exists ::env(XYMONNETWORK) ] } {
  unset ::env(XYMONNETWORK)
}
set XYMONHOSTGREP "[file join $XYMONHOME bin $XYMONSEARCH]"
#puts $OS
#puts $XYMON
#puts $XYMONCOMMS

set Options {
  {b          "Print configured Beans"}
  {n          "Dont send results to Xymon"}
  {r.arg  ""  "column name for status report"}
  {h.secret   "This Help"}
  {I.arg  ""  "Use the configuration file (default $XYMONHOME/etc/$HTAG.ini)"}
  {J.arg  ""  "Report specified JVM Name(s)"}
  {x.arg  ""  "Set debug level (0-9)"}
  {B          "Browse Mode"}
}

# Correct command-line invocation usage ....
set Command [::cmdline::getArgv0]
set Usage "Usage:\t$Command \[<options>\] \[<MBeans....>\]"
array set params [::cmdline::getoptions argv $Options $Usage]

if { $params(h) } {
  # error
  puts [::cmdline::usage $Options $Usage]
  exit 1
}

if { [string length $params(I)] > 0 } {
  set JMXINI $params(I)
} else {
  set JMXINI [ file join $XYMONHOME etc $HTAG.ini]
}
#puts $JMXINI
if { $params(x) > 0 } {
  set debug $params(x)
}
set report $HTAG
if { [string length $params(r)] > 0 } {
  if { [regexp -nocase {(none|z)} $params(r)] } {
    set report ""
  } else {
    set report $params(r)
  }
}
if { $params(n) } {
  if { $OS == "Windows" } {
    # we may be lucky ...
    set XYMON [ file join $XYMONHOME ext echo.exe]
  } else {
    set XYMON /bin/echo
  }
}

# if not in the c/l, select each host in hosts with this test defined
if { [string length $params(J)] > 0 } {
  set JavaInstances $params(J)
} else {
  if { $params(B) } {
    # error
    puts stderr "Must specify a Java Instance to use Browse Mode"
    puts stderr [::cmdline::usage $Options $Usage]
    exit 1
  }
  set JavaInstances ""
  set status [catch {exec $XYMONHOSTGREP --net $HTAG} result]
  if { $status == 0 } {
    # The command succeeded, and wrote nothing to stderr.
    # $result contains what it wrote to stdout
    set lines [split $result "\n"]
    foreach line $lines {
      lappend JavaInstances [lindex $line 1]
    }
  } elseif { [string equal $::errorCode NONE] } {
    # The command exited with a normal status, but wrote something
    # to stderr, which is included in $result.
    mydie $result
  } else {
    mydie "error executing $XYMONHOSTGREP --net $HTAG\n$result"
  }
}
#puts $JavaInstances

# this aborts if ini file cant be opened - configuration error
set myfh [ini::open $JMXINI r]

# these all abort if section or key cant be found - configuration error
if { [catch {string trim [ini::value $myfh default password]} defpassword] } {
  mydie "Problem in $JMXINI\n$defpassword"
}
if { [catch {string trim [ini::value $myfh default adminport]} defadminport] } {
  mydie "Problem in $JMXINI\n$defadminport"
}
if { [catch {string trim [ini::value $myfh default username]} defusername] } {
  mydie "Problem in $JMXINI\n$defusername"
}

# these all have inbuilt defaults - no configuration error
set timeoutcmdexec [string trim [ini::value $myfh default timeoutcmdexec 10]]
set conntest [string trim [ini::value $myfh default conntest "conn"]]

if { [llength $argv] } {
  # mbeans required from command line
  set datacollection $argv
} else {
  # inject some default mbeans here in case the default section is missing
  # from the ini file
  if { [ini::exists $myfh default datacollection] } {
    set defcollection [string trim [ini::value $myfh default datacollection]]
  } else {
    set defcollection [init_status datacollection]
  }
}

if { $params(b) } {
  if { [string length $params(J)] > 0 } {
    puts [string trim [ini::value $myfh $params(J) datacollection $defcollection]]
  } else {
    puts $defcollection
  }
  ini::close $myfh
  exit 0
}

set start [ clock click -milliseconds]
set beanstart [ clock click -milliseconds]
set hostschecked 0
set statusmsgs 0
set err 0
set warn 0
foreach {jvm} $JavaInstances {
  # for each jmx host...
  #
  # this aborts if section or key cant be found - configuration error
  if { [catch {string trim [ini::value $myfh $jvm hostname]} hostname] } {
    mydie "Problem in $JMXINI\nNo server found for jvm $jvm\n$hostname"
  }

  # defpassword can be blank
  if { [catch {string trim [ini::value $myfh $jvm password $defpassword]} password] } {
    set password $defpassword
  }

  # defuser can be blank
  if { [catch {string trim [ini::value $myfh $jvm username $defusername]} username] } {
    set username $defusername
  }

  # these inherit the default
  #set password [string trim [ini::value $myfh $jvm password $defpassword]]
  #set username [string trim [ini::value $myfh $jvm username $defusername]]
  set adminport [string trim [ini::value $myfh $jvm adminport $defadminport]]

  set thecreds ""
  if { [string length $username] > 0 } {
    set thecreds "-U $username"
  }
  if { [string length $password] > 0 } {
    set thecreds "$thecreds -P $password"
  }

  # see if Xymon thinks the host is reachable
  # no point trying to connect if not
  # if this fails, need to set all the dependant pages to clear
  set state [get_host_avail $hostname $conntest]

  # Connect to the JMX service if possible
  if { $state != $available } {

    # format is host column colour seconds, we have hijacked column
    set line [list $hostname "warning" "" ]
    lappend line "Host $hostname is down. Cannot Check JMX Status for JVM instance $jvm"
    S add row $line
    incr warn

  } else {

    if { $params(B) } {
      # Browse mode
      ini::close $myfh
      # this will retrieve all the mbeans in java.lang
      #set data [jmx_list java.lang:.*]
      # this doesnt work - would be nice if it retrieved all the attributes
  
      puts "-h $hostname -p $adminport -U $username -P $password"
      #exec sh -c {java -jar jmxsh-R5.jar -h $hostname -p $adminport -U $username -P $password} &
      #puts "finished"
      exit
    }

    # Set a timeout to make sure we get control back
    set state $disconnected
    set myafter [after 5000 set state $timeoutcmdexec]

    # defer the connection attempt till the wait is in place
    after 50 jmx_connect_proc -h $hostname -p $adminport $thecreds
    vwait state

    # wait is finished, clear up and see what we did get
    after cancel $myafter

    # set available 1
    # set connected 2
    # set disconnected 0
    # set timeoutcmdexec TIMEOUT
    # set failed FAILED
    switch -- $state {
      2 {
        debug_print "connected!"
      }
      0 {
        debug_print "not connected"
        set line [list $jvm "error" "" ]
        lappend line "JMX - no connection"
        S add row $line
        incr err
      }
      -1 {
        debug_print "timed out connecting on port $adminport"
        set line [list $jvm "error" "" ]
        lappend line "JMX - timed out connecting on port $adminport"
        S add row $line
        incr err
      }
      -2 {
        debug_print "connection failed"
        set line [list $jvm "error" "" ]
        lappend line "JMX - connection failed for JVM"
        S add row $line
        incr err
      }
    }
    if { $state == $connected } {
      if { [llength $argv] } {
        # mbeans required from command line
        set datacollection $argv
      } else {
        # we are looking for the mbeans in the ini file, if we are
        # connected OK then create the uptime Report page
        # This is a fake mbean which uses java.lang:type=Runtime
        # to fetch Uptime & StartTime
        #
        set mbean Uptime
        set attributes "StartTime Uptime"
        set column "Uptime"
        set beanstart [ clock click -milliseconds]
        set statusmsgs [expr $statusmsgs + [ jmx_lookup $jvm $mbean $column $attributes ]]

        # Fetch the list of remaining beans to be queried - inherits the default
        set datacollection [string trim [ini::value $myfh $jvm datacollection $defcollection]]
      }

      # step through each mbean creating Xymon Report pages as we go
      foreach {mbean} $datacollection {

        # this aborts if section or key cant be found - configuration error
        if { [catch {string trim [ini::value $myfh $mbean column]} column] } {
          mydie "Problem in $JMXINI\nNo column defined for mbean $mbean\n$column"
        }

        # this aborts if section or key cant be found - configuration error
        if { [catch {string trim [ini::value $myfh $mbean attributes]} attributes] } {
          mydie "Problem in $JMXINI\nNo attributes defined for mbean $mbean\n$attributes"
        }

        set beanstart [ clock click -milliseconds]
        set statusmsgs [expr $statusmsgs + [ jmx_lookup $jvm $mbean $column $attributes ]]
      }

      # Close this connection
      jmx_close
    }

  }
  # store some statistics
  set end [ clock click -milliseconds]
  set hostschecked [expr $hostschecked + 1]
  set elapsed [expr ( $end - $start ) / 1000]
  set line [list $jvm "" "" [expr {($end - $start)/1e3}]]
  S add row $line
  set start $end

}
ini::close $myfh
set end [ clock click -milliseconds]

if { [string length $report] > 0 } {
  # report ourselves
  set hostname [info hostname]
  debug_print "Creating $report summary page"
  set hostcomma [set_hostcomma $hostname ]
  if { $err } {
    set status "red"
  } elseif { $warn } {
    set status "yellow"
  } else {
    set status "green"
  }
  send_bb $hostcomma $report $status [print_data_full $hostname $report $hostschecked $statusmsgs]
  send_bb_data $hostcomma "trends" [format "\[%s.rrd\]\nDS:runtime:GAUGE:600:0:U %.2f\n" $report [expr {([ clock click -milliseconds] - $begin)/1e3}]]
}
S destroy
